from __future__ import annotations

from enum import Enum
from typing import List, Union

import numpy as np
import pandas as pd
import pyarrow as pa
from ray.data.dataset import Dataset
from daft import DataFrame as DaftDataFrame


LocalTable = Union[
    pa.Table,
    pd.DataFrame,
    np.ndarray,
    pa.parquet.ParquetFile,
]
LocalDataset = List[LocalTable]
DistributedDataset = Union[Dataset, DaftDataFrame]


class StreamFormat(str, Enum):
    DELTACAT = "deltacat"
    ICEBERG = "iceberg"
    HUDI = "hudi"
    DELTA_LAKE = "delta_lake"
    SQLITE3 = "SQLITE3"  # used by tests


class DeltaType(str, Enum):
    APPEND = "append"
    UPSERT = "upsert"
    DELETE = "delete"


class TransactionType(str, Enum):
    # the transaction reads existing data
    # does not conflict with any other transaction types
    READ = "read"
    # the transaction only appends new data
    # conflicts with other transaction types can be auto-resolved
    APPEND = "append"
    # the transaction alters existing data
    # (even if it also appends data)
    # conflicts with other alters/overwrites/restates/deletes fail
    ALTER = "alter"
    # the transaction overwrites existing data
    # (even if it also appends or alters data)
    # conflicts with other alters/overwrites/restates/deletes fail
    OVERWRITE = "overwrite"
    # the transaction restates existing data with a new layout
    # (even if it appends, alters, or overwrites data to do so)
    # conflicts with other alters/overwrites/restates/deletes fail
    RESTATE = "restate"
    # the transaction deletes existing data
    # (even if it also appends, alters, overwrites, or restates data)
    # conflicts with other alters/overwrites/restates/deletes fail
    DELETE = "delete"


class TransactionOperationType(str, Enum):
    CREATE = "create"
    UPDATE = "update"
    DELETE = "delete"

    READ_SIBLINGS = "read_siblings"
    READ_CHILDREN = "read_children"
    READ_LATEST = "read_latest"
    READ_EXISTS = "read_exists"

    @staticmethod
    def write_operations():
        return {
            TransactionOperationType.CREATE,
            TransactionOperationType.UPDATE,
            TransactionOperationType.DELETE,
        }

    @staticmethod
    def read_operations():
        return {
            TransactionOperationType.READ_SIBLINGS,
            TransactionOperationType.READ_CHILDREN,
            TransactionOperationType.READ_LATEST,
            TransactionOperationType.READ_EXISTS,
        }

    def is_write_operation(self) -> bool:
        return self in TransactionOperationType.write_operations()

    def is_read_operation(self) -> bool:
        return self in TransactionOperationType.read_operatins()


class LifecycleState(str, Enum):
    CREATED = "created"
    UNRELEASED = "unreleased"
    ACTIVE = "active"
    DEPRECATED = "deprecated"
    BETA = "beta"
    DELETED = "deleted"


class CommitState(str, Enum):
    STAGED = "staged"
    COMMITTED = "committed"
    DEPRECATED = "deprecated"


class SchemaConsistencyType(str, Enum):
    """
    DeltaCAT table schemas can be used to inform the data consistency checks
    run for each field. When present, the schema can be used to enforce the
    following field-level data consistency policies at table load time:

    NONE: No consistency checks are run.

    COERCE: Coerce fields to fit the schema whenever possible.

    VALIDATE: Raise an error for any fields that don't fit the schema.
    """

    NONE = "none"
    COERCE = "coerce"
    VALIDATE = "validate"


class SortOrder(str, Enum):
    ASCENDING = "ascending"
    DESCENDING = "descending"

    @classmethod
    def _missing_(cls, value: str):
        # pyiceberg.table.sorting.SortDirection mappings
        if value.lower() == "asc":
            return SortOrder.ASCENDING
        elif value.lower() == "desc":
            return SortOrder.DESCENDING
        return None


class NullOrder(str, Enum):
    AT_START = "at_start"
    AT_END = "at_end"

    @classmethod
    def _missing_(cls, value: str):
        # pyiceberg.table.sorting.NullOrder mappings
        if value.lower() == "nulls-first":
            return NullOrder.AT_START
        elif value.lower() == "nulls-last":
            return NullOrder.AT_END
        return None



from __future__ import annotations

import re
from typing import Optional, Tuple, Union, List

import sys
import urllib
import pathlib

import pyarrow
import pyarrow as pa
from pyarrow.fs import (
    _resolve_filesystem_and_path,
    FileSelector,
    FileInfo,
    FileType,
    FileSystem,
    FSSpecHandler,
    PyFileSystem,
)

_LOCAL_SCHEME = "local"


def resolve_paths_and_filesystem(
    paths: Union[str, List[str]],
    filesystem: pyarrow.fs.FileSystem = None,
) -> Tuple[List[str], pyarrow.fs.FileSystem]:
    """
    Resolves and normalizes all provided paths, infers a filesystem from the
    paths or validates the provided filesystem against the paths and ensures
    that all paths use the same filesystem.

    Args:
        paths: A single file/directory path or a list of file/directory paths.
            A list of paths can contain both files and directories.
        filesystem: The filesystem implementation that should be used for
            reading these files. If None, a filesystem will be inferred. If not
            None, the provided filesystem will still be validated against all
            filesystems inferred from the provided paths to ensure
            compatibility.
    """
    if isinstance(paths, str):
        paths = [paths]
    if isinstance(paths, pathlib.Path):
        paths = [str(paths)]
    elif not isinstance(paths, list) or any(not isinstance(p, str) for p in paths):
        raise ValueError(
            "Expected `paths` to be a `str`, `pathlib.Path`, or `list[str]`, but got "
            f"`{paths}`."
        )
    elif len(paths) == 0:
        raise ValueError("Must provide at least one path.")

    need_unwrap_path_protocol = True
    if filesystem and not isinstance(filesystem, FileSystem):
        err_msg = (
            f"The filesystem passed must either conform to "
            f"pyarrow.fs.FileSystem, or "
            f"fsspec.spec.AbstractFileSystem. The provided "
            f"filesystem was: {filesystem}"
        )
        try:
            import fsspec
            from fsspec.implementations.http import HTTPFileSystem
        except ModuleNotFoundError:
            # If filesystem is not a pyarrow filesystem and fsspec isn't
            # installed, then filesystem is neither a pyarrow filesystem nor
            # an fsspec filesystem, so we raise a TypeError.
            raise TypeError(err_msg) from None
        if not isinstance(filesystem, fsspec.spec.AbstractFileSystem):
            raise TypeError(err_msg) from None
        if isinstance(filesystem, HTTPFileSystem):
            # If filesystem is fsspec HTTPFileSystem, the protocol/scheme of paths
            # should not be unwrapped/removed, because HTTPFileSystem expects full file
            # paths including protocol/scheme. This is different behavior compared to
            # file systems implementation in pyarrow.fs.FileSystem.
            need_unwrap_path_protocol = False

        filesystem = PyFileSystem(FSSpecHandler(filesystem))

    resolved_paths = []
    for path in paths:
        path = _resolve_custom_scheme(path)
        try:
            resolved_filesystem, resolved_path = _resolve_filesystem_and_path(
                path, filesystem
            )
        except pa.lib.ArrowInvalid as e:
            if "Cannot parse URI" in str(e):
                resolved_filesystem, resolved_path = _resolve_filesystem_and_path(
                    _encode_url(path), filesystem
                )
                resolved_path = _decode_url(resolved_path)
            elif "Unrecognized filesystem type in URI" in str(e):
                scheme = urllib.parse.urlparse(path, allow_fragments=False).scheme
                if scheme in ["http", "https"]:
                    # If scheme of path is HTTP and filesystem is not resolved,
                    # try to use fsspec HTTPFileSystem. This expects fsspec is
                    # installed.
                    try:
                        from fsspec.implementations.http import HTTPFileSystem
                    except ModuleNotFoundError:
                        raise ImportError(
                            "Please install fsspec to read files from HTTP."
                        ) from None

                    resolved_filesystem = PyFileSystem(FSSpecHandler(HTTPFileSystem()))
                    resolved_path = path
                    need_unwrap_path_protocol = False
                else:
                    raise
            else:
                raise
        if filesystem is None:
            filesystem = resolved_filesystem
        elif need_unwrap_path_protocol:
            resolved_path = _unwrap_protocol(resolved_path)
        resolved_path = filesystem.normalize_path(resolved_path)
        resolved_paths.append(resolved_path)

    return resolved_paths, filesystem


def resolve_path_and_filesystem(
    path: str,
    filesystem: Optional[pyarrow.fs.FileSystem] = None,
) -> Tuple[str, pyarrow.fs.FileSystem]:
    """
    Resolves and normalizes the provided path, infers a filesystem from the
    path or validates the provided filesystem against the path.

    Args:
        path: A single file/directory path.
        filesystem: The filesystem implementation that should be used for
            reading these files. If None, a filesystem will be inferred. If not
            None, the provided filesystem will still be validated against all
            filesystems inferred from the provided paths to ensure
            compatibility.
    """
    paths, filesystem = resolve_paths_and_filesystem(
        paths=path,
        filesystem=filesystem,
    )
    assert len(paths) == 1, len(paths)
    return paths[0], filesystem


def list_directory(
    path: str,
    filesystem: pyarrow.fs.FileSystem,
    exclude_prefixes: Optional[List[str]] = None,
    ignore_missing_path: bool = False,
    recursive: bool = False,
) -> List[Tuple[str, int]]:
    """
    Expand the provided directory path to a list of file paths.

    Args:
        path: The directory path to expand.
        filesystem: The filesystem implementation that should be used for
            reading these files.
        exclude_prefixes: The file relative path prefixes that should be
            excluded from the returned file set. Default excluded prefixes are
            "." and "_".
        recursive: Whether to expand subdirectories or not.

    Returns:
        An iterator of (file_path, file_size) tuples.
    """
    if exclude_prefixes is None:
        exclude_prefixes = [".", "_"]

    selector = FileSelector(
        base_dir=path,
        recursive=recursive,
        allow_not_found=ignore_missing_path,
    )
    try:
        files = filesystem.get_file_info(selector)
    except OSError as e:
        if isinstance(e, FileNotFoundError):
            files = []
        else:
            _handle_read_os_error(e, path)
    base_path = selector.base_dir
    out = []
    for file_ in files:
        file_path = file_.path
        if not file_path.startswith(base_path):
            continue
        relative = file_path[len(base_path) :]
        if any(relative.startswith(prefix) for prefix in exclude_prefixes):
            continue
        out.append((file_path, file_.size))
    # We sort the paths to guarantee a stable order.
    return sorted(out)


def get_file_info(
    path: str,
    filesystem: pyarrow.fs.FileSystem,
    ignore_missing_path: bool = False,
) -> FileInfo:
    """Get the file info for the provided path."""
    try:
        file_info = filesystem.get_file_info(path)
    except OSError as e:
        _handle_read_os_error(e, path)
    if file_info.type == FileType.NotFound and not ignore_missing_path:
        raise FileNotFoundError(path)

    return file_info


def _handle_read_os_error(
    error: OSError,
    paths: Union[str, List[str]],
) -> str:
    # NOTE: this is not comprehensive yet, and should be extended as more errors arise.
    # NOTE: The latter patterns are raised in Arrow 10+, while the former is raised in
    # Arrow < 10.
    aws_error_pattern = (
        r"^(?:(.*)AWS Error \[code \d+\]: No response body\.(.*))|"
        r"(?:(.*)AWS Error UNKNOWN \(HTTP status 400\) during HeadObject operation: "
        r"No response body\.(.*))|"
        r"(?:(.*)AWS Error ACCESS_DENIED during HeadObject operation: No response "
        r"body\.(.*))$"
    )
    if re.match(aws_error_pattern, str(error)):
        # Specially handle AWS error when reading files, to give a clearer error
        # message to avoid confusing users. The real issue is most likely that the AWS
        # S3 file credentials have not been properly configured yet.
        if isinstance(paths, str):
            # Quote to highlight single file path in error message for better
            # readability. List of file paths will be shown up as ['foo', 'boo'],
            # so only quote single file path here.
            paths = f'"{paths}"'
        raise OSError(
            (
                f"Failing to read AWS S3 file(s): {paths}. "
                "Please check that file exists and has properly configured access. "
                "You can also run AWS CLI command to get more detailed error message "
                "(e.g., aws s3 ls <file-name>). "
                "See https://awscli.amazonaws.com/v2/documentation/api/latest/reference/s3/index.html "  # noqa
                "for more information."
            )
        )
    else:
        raise error


def _is_local_windows_path(path: str) -> bool:
    """Determines if path is a Windows file-system location."""
    if sys.platform != "win32":
        return False

    if len(path) >= 1 and path[0] == "\\":
        return True
    if (
        len(path) >= 3
        and path[1] == ":"
        and (path[2] == "/" or path[2] == "\\")
        and path[0].isalpha()
    ):
        return True
    return False


def _unwrap_protocol(path):
    """
    Slice off any protocol prefixes on path.
    """
    if sys.platform == "win32" and _is_local_windows_path(path):
        # Represent as posix path such that downstream functions properly handle it.
        # This is executed when 'file://' is NOT included in the path.
        return pathlib.Path(path).as_posix()

    parsed = urllib.parse.urlparse(path, allow_fragments=False)  # support '#' in path
    query = "?" + parsed.query if parsed.query else ""  # support '?' in path
    netloc = parsed.netloc
    if parsed.scheme == "s3" and "@" in parsed.netloc:
        # If the path contains an @, it is assumed to be an anonymous
        # credentialed path, and we need to strip off the credentials.
        netloc = parsed.netloc.split("@")[-1]

    parsed_path = parsed.path
    # urlparse prepends the path with a '/'. This does not work on Windows
    # so if this is the case strip the leading slash.
    if (
        sys.platform == "win32"
        and not netloc
        and len(parsed_path) >= 3
        and parsed_path[0] == "/"  # The problematic leading slash
        and parsed_path[1].isalpha()  # Ensure it is a drive letter.
        and parsed_path[2:4] in (":", ":/")
    ):
        parsed_path = parsed_path[1:]

    return netloc + parsed_path + query


def _encode_url(path):
    return urllib.parse.quote(path, safe="/:")


def _decode_url(path):
    return urllib.parse.unquote(path)


def _resolve_custom_scheme(path: str) -> str:
    """Returns the resolved path if the given path follows a Ray-specific custom
    scheme. Othewise, returns the path unchanged.

    The supported custom schemes are: "local", "example".
    """
    parsed_uri = urllib.parse.urlparse(path)
    if parsed_uri.scheme == _LOCAL_SCHEME:
        path = parsed_uri.netloc + parsed_uri.path
    return path



import uuid

from typing import Any, Callable, Dict, List, Optional, Union, Tuple

from deltacat.catalog import get_catalog_properties
from deltacat.constants import DEFAULT_TABLE_VERSION
from deltacat.exceptions import TableNotFoundError
from deltacat.storage.model.manifest import (
    EntryParams,
    ManifestAuthor,
)
from deltacat.storage.model.delta import (
    Delta,
    DeltaLocator,
    DeltaProperties,
    DeltaType,
)
from deltacat.storage.model.types import (
    CommitState,
    DistributedDataset,
    LifecycleState,
    LocalDataset,
    LocalTable,
    TransactionType,
    TransactionOperationType,
    StreamFormat,
)
from deltacat.storage.model.list_result import ListResult
from deltacat.storage.model.namespace import (
    Namespace,
    NamespaceLocator,
    NamespaceProperties,
)
from deltacat.storage.model.partition import (
    Partition,
    PartitionLocator,
    PartitionScheme,
    PartitionValues,
    UNPARTITIONED_SCHEME_ID,
    PartitionLocatorAlias,
)
from deltacat.storage.model.schema import (
    Schema,
)
from deltacat.storage.model.sort_key import (
    SortScheme,
)
from deltacat.storage.model.stream import (
    Stream,
    StreamLocator,
)
from deltacat.storage.model.table import (
    Table,
    TableProperties,
    TableLocator,
)
from deltacat.storage.model.table_version import (
    TableVersion,
    TableVersionProperties,
    TableVersionLocator,
)
from deltacat.storage.model.metafile import (
    Metafile,
)
from deltacat.storage.model.transaction import (
    TransactionOperation,
    Transaction,
    TransactionOperationList,
)
from deltacat.storage.model.manifest import Manifest
from deltacat.types.media import (
    ContentType,
    DistributedDatasetType,
    StorageType,
    TableType,
)
from deltacat.utils.common import ReadKwargsProvider


def _list(
    metafile: Metafile,
    txn_op_type: TransactionOperationType,
    *args,
    **kwargs,
) -> ListResult[Metafile]:
    catalog_properties = get_catalog_properties(**kwargs)
    limit = kwargs.get("limit") or None
    transaction = Transaction.of(
        txn_type=TransactionType.READ,
        txn_operations=[
            TransactionOperation.of(
                operation_type=txn_op_type,
                dest_metafile=metafile,
                read_limit=limit,
            )
        ],
    )
    list_results_per_op = transaction.commit(
        catalog_root_dir=catalog_properties.root,
        filesystem=catalog_properties.filesystem,
    )
    return list_results_per_op[0]


def _latest(
    metafile: Metafile,
    *args,
    **kwargs,
) -> Optional[Metafile]:
    list_results = _list(
        *args,
        metafile=metafile,
        txn_op_type=TransactionOperationType.READ_LATEST,
        **kwargs,
    )
    results = list_results.all_items()
    return results[0] if results else None


def _exists(
    metafile: Metafile,
    *args,
    **kwargs,
) -> Optional[Metafile]:
    list_results = _list(
        *args,
        metafile=metafile,
        txn_op_type=TransactionOperationType.READ_EXISTS,
        **kwargs,
    )
    results = list_results.all_items()
    return True if results else False


def _resolve_partition_locator_alias(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    partition_values: Optional[PartitionValues] = None,
    partition_scheme_id: Optional[str] = None,
    *args,
    **kwargs,
) -> PartitionLocatorAlias:
    # TODO(pdames): A read shouldn't initiate N transactions that
    #  read against different catalog snapshots. To resolve this, add
    #  new "start", "step", and "end" methods to Transaction that
    #  support starting a txn, defining and executing a txn op, retrieve
    #  its results, then define and execute the next txn op. When
    #  stepping through a transaction its txn heartbeat timeout should
    #  be set manually.
    partition_locator = None
    if not partition_values:
        partition_scheme_id = UNPARTITIONED_SCHEME_ID
    elif not partition_scheme_id:
        # resolve latest partition scheme from the current
        # revision of its `deltacat` stream
        stream = get_stream(
            *args,
            namespace=namespace,
            table_name=table_name,
            table_version=table_version,
            **kwargs,
        )
        if not stream:
            raise ValueError(
                f"Failed to resolve latest partition scheme for "
                f"`{namespace}.{table_name}` at table version "
                f"`{table_version or 'latest'}` (no stream found)."
            )
        partition_locator = PartitionLocator.of(
            stream_locator=stream.locator,
            partition_values=partition_values,
            partition_id=None,
        )
        partition_scheme_id = stream.partition_scheme.id
    if not partition_locator:
        partition_locator = PartitionLocator.at(
            namespace=namespace,
            table_name=table_name,
            table_version=table_version,
            stream_id=None,
            stream_format=StreamFormat.DELTACAT,
            partition_values=partition_values,
            partition_id=None,
        )
    partition = Partition.of(
        locator=partition_locator,
        schema=None,
        content_types=None,
        partition_scheme_id=partition_scheme_id,
    )
    return partition.locator_alias


def _resolve_latest_active_table_version_id(
    namespace: str,
    table_name: str,
    fail_if_no_active_table_version: True,
    *args,
    **kwargs,
) -> Optional[str]:
    table = get_table(
        *args,
        namespace=namespace,
        table_name=table_name,
        **kwargs,
    )
    if not table:
        raise ValueError(f"Table does not exist: {namespace}.{table_name}")
    if fail_if_no_active_table_version and not table.latest_active_table_version:
        raise ValueError(f"Table has no active table version: {namespace}.{table_name}")
    return table.latest_active_table_version


def _resolve_latest_table_version_id(
    namespace: str,
    table_name: str,
    fail_if_no_active_table_version: True,
    *args,
    **kwargs,
) -> Optional[str]:
    table = get_table(
        *args,
        namespace=namespace,
        table_name=table_name,
        **kwargs,
    )
    if not table:
        raise ValueError(f"Table does not exist: {namespace}.{table_name}")
    if fail_if_no_active_table_version and not table.latest_table_version:
        raise ValueError(f"Table has no table version: {namespace}.{table_name}")
    return table.latest_table_version


def list_namespaces(*args, **kwargs) -> ListResult[Namespace]:
    """
    Lists a page of table namespaces. Namespaces are returned as list result
    items.
    """
    return _list(
        *args,
        metafile=Namespace.of(NamespaceLocator.of("placeholder")),
        txn_op_type=TransactionOperationType.READ_SIBLINGS,
        **kwargs,
    )


def list_tables(namespace: str, *args, **kwargs) -> ListResult[Table]:
    """
    Lists a page of tables for the given table namespace. Tables are returned as
    list result items. Raises an error if the given namespace does not exist.
    """
    locator = TableLocator.at(namespace=namespace, table_name="placeholder")
    return _list(
        *args,
        metafile=Table.of(locator=locator),
        txn_op_type=TransactionOperationType.READ_SIBLINGS,
        **kwargs,
    )


def list_table_versions(
    namespace: str,
    table_name: str,
    *args,
    **kwargs,
) -> ListResult[TableVersion]:
    """
    Lists a page of table versions for the given table. Table versions are
    returned as list result items. Raises an error if the given table does not
    exist.
    """
    locator = TableVersionLocator.at(
        namespace=namespace,
        table_name=table_name,
        table_version="placeholder.0",
    )
    table_version = TableVersion.of(
        locator=locator,
        schema=None,
    )
    return _list(
        *args,
        metafile=table_version,
        txn_op_type=TransactionOperationType.READ_SIBLINGS,
        **kwargs,
    )


def list_streams(
    namespace: str,
    table_name: str,
    table_version: str,
    *args,
    **kwargs,
) -> ListResult[Stream]:
    """
    Lists a page of streams for the given table version.
    Raises an error if the table version does not exist.
    """
    locator = StreamLocator.at(
        namespace=namespace,
        table_name=table_name,
        table_version=table_version,
        stream_id="placeholder",
        stream_format=None,
    )
    stream = Stream.of(
        locator=locator,
        partition_scheme=None,
    )
    return _list(
        stream,
        TransactionOperationType.READ_SIBLINGS,
        *args,
        **kwargs,
    )


def list_partitions(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    *args,
    **kwargs,
) -> ListResult[Partition]:
    """
    Lists a page of partitions for the given table version. Partitions are
    returned as list result items. Table version resolves to the latest active
    table version if not specified. Raises an error if the table version does
    not exist.
    """
    locator = PartitionLocator.at(
        namespace=namespace,
        table_name=table_name,
        table_version=table_version,
        stream_id=None,
        stream_format=StreamFormat.DELTACAT,
        partition_values=["placeholder"],
        partition_id="placeholder",
    )
    partition = Partition.of(
        locator=locator,
        schema=None,
        content_types=None,
    )
    return _list(
        *args,
        metafile=partition,
        txn_op_type=TransactionOperationType.READ_SIBLINGS,
        **kwargs,
    )


def list_stream_partitions(stream: Stream, *args, **kwargs) -> ListResult[Partition]:
    """
    Lists all partitions committed to the given stream.
    """
    if stream.stream_format != StreamFormat.DELTACAT:
        raise ValueError(
            f"Unsupported stream format: {stream.stream_format}"
            f"Expected stream format: {StreamFormat.DELTACAT}"
        )
    locator = PartitionLocator.of(
        stream_locator=stream.locator,
        partition_values=["placeholder"],
        partition_id="placeholder",
    )
    partition = Partition.of(
        locator=locator,
        schema=None,
        content_types=None,
    )
    return _list(
        *args,
        metafile=partition,
        txn_op_type=TransactionOperationType.READ_SIBLINGS,
        **kwargs,
    )


def list_deltas(
    namespace: str,
    table_name: str,
    partition_values: Optional[PartitionValues] = None,
    table_version: Optional[str] = None,
    first_stream_position: Optional[int] = None,
    last_stream_position: Optional[int] = None,
    ascending_order: Optional[bool] = None,
    include_manifest: bool = False,
    partition_scheme_id: Optional[str] = None,
    *args,
    **kwargs,
) -> ListResult[Delta]:
    """
    Lists a page of deltas for the given table version and committed partition.
    Deltas are returned as list result items. Deltas returned can optionally be
    limited to inclusive first and last stream positions. Deltas are returned by
    descending stream position by default. Table version resolves to the latest
    active table version if not specified. Partition values should not be
    specified for unpartitioned tables. Partition scheme ID resolves to the
    table version's current partition scheme by default. Raises an error if the
    given table version or partition does not exist.

    To conserve memory, the deltas returned do not include manifests by
    default. The manifests can either be optionally retrieved as part of this
    call or lazily loaded via subsequent calls to `get_delta_manifest`.
    """
    # TODO(pdames): Delta listing should ideally either use an efficient
    #  range-limited dir listing of partition children between start and end
    #  positions, or should traverse using Partition.stream_position (to
    #  resolve last stream position) and Delta.previous_stream_position
    #  (down to first stream position).
    partition_locator_alias = _resolve_partition_locator_alias(
        *args,
        namespace=namespace,
        table_name=table_name,
        table_version=table_version,
        partition_values=partition_values,
        partition_scheme_id=partition_scheme_id,
        **kwargs,
    )
    locator = DeltaLocator.of(locator=partition_locator_alias)
    delta = Delta.of(
        locator=locator,
        delta_type=None,
        meta=None,
        properties=None,
        manifest=None,
    )
    all_deltas_list_result: ListResult[Delta] = _list(
        *args,
        metafile=delta,
        txn_op_type=TransactionOperationType.READ_SIBLINGS,
        **kwargs,
    )
    all_deltas = all_deltas_list_result.all_items()
    filtered_deltas = [
        delta
        for delta in all_deltas
        if first_stream_position <= delta.stream_position <= last_stream_position
    ]
    if ascending_order:
        filtered_deltas.reverse()
    return filtered_deltas


def list_partition_deltas(
    partition_like: Union[Partition, PartitionLocator],
    first_stream_position: Optional[int] = None,
    last_stream_position: Optional[int] = None,
    ascending_order: bool = False,
    include_manifest: bool = False,
    *args,
    **kwargs,
) -> ListResult[Delta]:
    """
    Lists a page of deltas committed to the given partition.

    To conserve memory, the deltas returned do not include manifests by
    default. The manifests can either be optionally retrieved as part of this
    call or lazily loaded via subsequent calls to `get_delta_manifest`.
    """
    # TODO(pdames): Delta listing should ideally either use an efficient
    #  range-limited dir listing of partition children between start and end
    #  positions, or should traverse using Partition.stream_position (to
    #  resolve last stream position) and Delta.previous_stream_position
    #  (down to first stream position).
    locator = DeltaLocator.of(
        partition_locator=partition_like
        if isinstance(partition_like, PartitionLocator)
        else partition_like.locator,
        stream_position=None,
    )
    delta = Delta.of(
        locator=locator,
        delta_type=None,
        meta=None,
        properties=None,
        manifest=None,
    )
    all_deltas_list_result: ListResult[Delta] = _list(
        *args,
        metafile=delta,
        txn_op_type=TransactionOperationType.READ_SIBLINGS,
        **kwargs,
    )
    all_deltas = all_deltas_list_result.all_items()
    filtered_deltas = [
        delta
        for delta in all_deltas
        if first_stream_position <= delta.stream_position <= last_stream_position
    ]
    if ascending_order:
        filtered_deltas.reverse()
    return filtered_deltas


def get_delta(
    namespace: str,
    table_name: str,
    stream_position: int,
    partition_values: Optional[PartitionValues] = None,
    table_version: Optional[str] = None,
    include_manifest: bool = False,
    partition_scheme_id: Optional[str] = None,
    *args,
    **kwargs,
) -> Optional[Delta]:
    """
    Gets the delta for the given table version, partition, and stream position.
    Table version resolves to the latest active table version if not specified.
    Partition values should not be specified for unpartitioned tables. Partition
    scheme ID resolves to the table version's current partition scheme by
    default. Raises an error if the given table version or partition does not
    exist.

    To conserve memory, the delta returned does not include a manifest by
    default. The manifest can either be optionally retrieved as part of this
    call or lazily loaded via a subsequent call to `get_delta_manifest`.
    """
    # TODO(pdames): Honor `include_manifest` param.
    partition_locator_alias = _resolve_partition_locator_alias(
        *args,
        namespace=namespace,
        table_name=table_name,
        table_version=table_version,
        partition_values=partition_values,
        partition_scheme_id=partition_scheme_id,
        **kwargs,
    )
    locator = DeltaLocator.of(
        locator=partition_locator_alias,
        stream_position=stream_position,
    )
    delta = Delta.of(
        locator=locator,
        delta_type=None,
        meta=None,
        properties=None,
        manifest=None,
    )
    return _latest(
        *args,
        metafile=delta,
        **kwargs,
    )


def get_latest_delta(
    namespace: str,
    table_name: str,
    partition_values: Optional[PartitionValues] = None,
    table_version: Optional[str] = None,
    include_manifest: bool = False,
    partition_scheme_id: Optional[str] = None,
    *args,
    **kwargs,
) -> Optional[Delta]:
    """
    Gets the latest delta (i.e. the delta with the greatest stream position) for
    the given table version and partition. Table version resolves to the latest
    active table version if not specified. Partition values should not be
    specified for unpartitioned tables. Partition scheme ID resolves to the
    table version's current partition scheme by default. Raises an error if the
    given table version or partition does not exist.

    To conserve memory, the delta returned does not include a manifest by
    default. The manifest can either be optionally retrieved as part of this
    call or lazily loaded via a subsequent call to `get_delta_manifest`.
    """
    # TODO(pdames): Wrap this method in 1 single txn.
    stream = get_stream(
        namespace=namespace,
        table_name=table_name,
        table_version=table_version,
    )
    partition = get_partition(
        stream_locator=stream.locator,
        partition_values=partition_values,
        partition_scheme_id=partition_scheme_id,
    )
    locator = DeltaLocator.of(
        locator=partition.locator,
        stream_position=partition.stream_position,
    )
    delta = Delta.of(
        locator=locator,
        delta_type=None,
        meta=None,
        properties=None,
        manifest=None,
    )
    return _latest(
        *args,
        metafile=delta,
        **kwargs,
    )


def download_delta(
    delta_like: Union[Delta, DeltaLocator],
    table_type: TableType = TableType.PYARROW,
    storage_type: StorageType = StorageType.DISTRIBUTED,
    max_parallelism: Optional[int] = None,
    columns: Optional[List[str]] = None,
    file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
    ray_options_provider: Callable[[int, Any], Dict[str, Any]] = None,
    distributed_dataset_type: DistributedDatasetType = DistributedDatasetType.RAY_DATASET,
    *args,
    **kwargs,
) -> Union[LocalDataset, DistributedDataset]:  # type: ignore
    """
    Download the given delta or delta locator into either a list of
    tables resident in the local node's memory, or into a dataset distributed
    across this Ray cluster's object store memory. Ordered table N of a local
    table list, or ordered block N of a distributed dataset, always contain
    the contents of ordered delta manifest entry N.
    """
    raise NotImplementedError("download_delta not implemented")


def download_delta_manifest_entry(
    delta_like: Union[Delta, DeltaLocator],
    entry_index: int,
    table_type: TableType = TableType.PYARROW,
    columns: Optional[List[str]] = None,
    file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
    *args,
    **kwargs,
) -> LocalTable:
    """
    Downloads a single manifest entry into the specified table type for the
    given delta or delta locator. If a delta is provided with a non-empty
    manifest, then the entry is downloaded from this manifest. Otherwise, the
    manifest is first retrieved then the given entry index downloaded.

    NOTE: The entry will be downloaded in the current node's memory.
    """
    raise NotImplementedError("download_delta_manifest_entry not implemented")


def get_delta_manifest(
    delta_like: Union[Delta, DeltaLocator],
    *args,
    **kwargs,
) -> Manifest:
    """
    Get the manifest associated with the given delta or delta locator. This
    always retrieves the authoritative durable copy of the delta manifest, and
    never the local manifest defined for any input delta. Raises an error if
    the delta can't be found, or if it doesn't contain a manifest.
    """
    if isinstance(delta_like, Delta):
        delta_locator = delta_like.locator
    elif isinstance(delta_like, DeltaLocator):
        delta_locator = delta_like
    else:
        raise ValueError(
            f"Expected delta or delta locator, but got: {type(delta_like)}"
        )
    delta = Delta.of(
        locator=delta_locator,
        delta_type=None,
        meta=None,
        properties=None,
        manifest=None,
    )
    latest_delta = _latest(
        metafile=delta,
        *args,
        **kwargs,
    )
    if not latest_delta or not latest_delta.manifest:
        raise ValueError(f"No manifest found for delta: {delta_locator}")
    return latest_delta.manifest


def create_namespace(
    namespace: str,
    properties: Optional[NamespaceProperties] = None,
    *args,
    **kwargs,
) -> Namespace:
    """
    Creates a table namespace with the given name and properties. Returns
    the created namespace.
    """
    namespace = Namespace.of(
        locator=NamespaceLocator.of(namespace=namespace),
        properties=properties,
    )
    transaction = Transaction.of(
        txn_type=TransactionType.APPEND,
        txn_operations=[
            TransactionOperation.of(
                operation_type=TransactionOperationType.CREATE,
                dest_metafile=namespace,
            )
        ],
    )
    catalog_properties = get_catalog_properties(**kwargs)
    transaction.commit(
        catalog_root_dir=catalog_properties.root,
        filesystem=catalog_properties.filesystem,
    )
    return namespace


def update_namespace(
    namespace: str,
    properties: Optional[NamespaceProperties] = None,
    new_namespace: Optional[str] = None,
    *args,
    **kwargs,
) -> None:
    """
    Updates a table namespace's name and/or properties. Raises an error if the
    given namespace does not exist.
    """
    # TODO(pdames): Wrap get & update within a single txn.
    old_namespace = get_namespace(
        *args,
        namespace=namespace,
        **kwargs,
    )
    new_namespace: Namespace = Metafile.update_for(old_namespace)
    new_namespace.namespace = namespace
    new_namespace.properties = properties
    transaction = Transaction.of(
        txn_type=TransactionType.ALTER,
        txn_operations=[
            TransactionOperation.of(
                operation_type=TransactionOperationType.UPDATE,
                dest_metafile=new_namespace,
                src_metafile=old_namespace,
            )
        ],
    )
    catalog_properties = get_catalog_properties(**kwargs)
    transaction.commit(
        catalog_root_dir=catalog_properties.root,
        filesystem=catalog_properties.filesystem,
    )
    return namespace


def create_table_version(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    schema: Optional[Schema] = None,
    partition_scheme: Optional[PartitionScheme] = None,
    sort_keys: Optional[SortScheme] = None,
    table_version_description: Optional[str] = None,
    table_version_properties: Optional[TableVersionProperties] = None,
    table_description: Optional[str] = None,
    table_properties: Optional[TableProperties] = None,
    supported_content_types: Optional[List[ContentType]] = None,
    *args,
    **kwargs,
) -> Tuple[Table, TableVersion, Stream]:
    """
    Create a table version with an unreleased lifecycle state and an empty delta
    stream. Table versions may be schemaless and unpartitioned to improve write
    performance, or have their writes governed by a schema and partition scheme
    to improve data consistency and read performance.

    Returns a tuple containing the created/updated table, table version, and
    stream (respectively).

    Raises an error if the given namespace does not exist.
    """
    if not namespace_exists(
        *args,
        namespace=namespace,
        **kwargs,
    ):
        raise ValueError(f"Namespace {namespace} does not exist")
    # check if a parent table and/or previous table version already exist
    prev_table_version = None
    prev_table = get_table(
        *args,
        namespace=namespace,
        table_name=table_name,
        **kwargs,
    )
    if not prev_table:
        # no parent table exists, so we'll create it in this transaction
        txn_type = TransactionType.APPEND
        table_txn_op_type = TransactionOperationType.CREATE
        prev_table = None
        new_table = Table.of(
            locator=TableLocator.at(namespace=namespace, table_name=table_name),
        )
        table_version = table_version or DEFAULT_TABLE_VERSION
    else:
        # the parent table exists, so we'll update it in this transaction
        txn_type = TransactionType.ALTER
        table_txn_op_type = TransactionOperationType.UPDATE
        new_table: Table = Metafile.update_for(prev_table)
        prev_table_version = prev_table.latest_table_version
        if not table_version:
            # generate the next table version ID
            table_version = TableVersion.next_version(prev_table_version)
        else:
            # ensure that the given table version number matches expectations
            expected_table_version = TableVersion.next_version(prev_table_version)
            _, version_number = TableVersion.parse_table_version(
                table_version,
            )
            _, expected_version_number = TableVersion.parse_table_version(
                expected_table_version,
            )
            if version_number != expected_version_number:
                raise ValueError(
                    f"Expected to create table version "
                    f"{expected_version_number} but found {version_number}.",
                )
    new_table.description = table_description or table_version_description
    new_table.properties = table_properties
    new_table.latest_table_version = table_version
    catalog_properties = get_catalog_properties(**kwargs)
    locator = TableVersionLocator.at(
        namespace=namespace,
        table_name=table_name,
        table_version=table_version,
    )
    table_version = TableVersion.of(
        locator=locator,
        schema=schema,
        partition_scheme=partition_scheme,
        description=table_version_description,
        properties=table_version_properties,
        content_types=supported_content_types,
        sort_scheme=sort_keys,
        watermark=None,
        lifecycle_state=LifecycleState.CREATED,
        schemas=[schema] if schema else None,
        partition_schemes=[partition_scheme] if partition_scheme else None,
        sort_schemes=[sort_keys] if sort_keys else None,
        previous_table_version=prev_table_version,
    )
    # create the table version's default deltacat stream in this transaction
    stream_locator = StreamLocator.of(
        table_version_locator=locator,
        stream_id=str(uuid.uuid4()),
        stream_format=StreamFormat.DELTACAT,
    )
    stream = Stream.of(
        locator=stream_locator,
        partition_scheme=partition_scheme,
        state=CommitState.COMMITTED,
        previous_stream_id=None,
        watermark=None,
    )
    transaction = Transaction.of(
        txn_type=txn_type,
        txn_operations=[
            TransactionOperation.of(
                operation_type=table_txn_op_type,
                dest_metafile=new_table,
                src_metafile=prev_table,
            ),
            TransactionOperation.of(
                operation_type=TransactionOperationType.CREATE,
                dest_metafile=table_version,
            ),
            TransactionOperation.of(
                operation_type=TransactionOperationType.CREATE,
                dest_metafile=stream,
            ),
        ],
    )
    transaction.commit(
        catalog_root_dir=catalog_properties.root,
        filesystem=catalog_properties.filesystem,
    )
    return new_table, table_version, stream


def update_table(
    namespace: str,
    table_name: str,
    description: Optional[str] = None,
    properties: Optional[TableProperties] = None,
    new_table_name: Optional[str] = None,
    *args,
    **kwargs,
) -> None:
    """
    Update table metadata describing the table versions it contains. By default,
    a table's properties are empty, and its description is equal to that given
    when its first table version was created. Raises an error if the given
    table does not exist.
    """
    old_table = get_table(
        *args,
        namespace=namespace,
        table_name=table_name,
        **kwargs,
    )
    if not old_table:
        raise TableNotFoundError(f"Table `{namespace}.{table_name}` does not exist.")
    new_table: Table = Metafile.update_for(old_table)
    new_table.description = description or old_table.description
    new_table.properties = properties or old_table.properties
    new_table.table_name = new_table_name or old_table.table_name
    transaction = Transaction.of(
        txn_type=TransactionType.ALTER,
        txn_operations=[
            TransactionOperation.of(
                operation_type=TransactionOperationType.UPDATE,
                dest_metafile=new_table,
                src_metafile=old_table,
            )
        ],
    )
    catalog_properties = get_catalog_properties(**kwargs)
    transaction.commit(
        catalog_root_dir=catalog_properties.root,
        filesystem=catalog_properties.filesystem,
    )


def update_table_version(
    namespace: str,
    table_name: str,
    table_version: str,
    lifecycle_state: Optional[LifecycleState] = None,
    schema: Optional[Schema] = None,
    description: Optional[str] = None,
    properties: Optional[TableVersionProperties] = None,
    partition_scheme: Optional[PartitionScheme] = None,
    sort_keys: Optional[SortScheme] = None,
    *args,
    **kwargs,
) -> None:
    """
    Update a table version. Notably, updating an unreleased table version's
    lifecycle state to 'active' telegraphs that it is ready for external
    consumption, and causes all calls made to consume/produce streams,
    partitions, or deltas from/to its parent table to automatically resolve to
    this table version by default (i.e., when the client does not explicitly
    specify a different table version). Raises an error if the given table
    version does not exist.
    """
    # TODO(pdames): Wrap get & update within a single txn.
    old_table_version = get_table_version(
        *args,
        namespace=namespace,
        table_name=table_name,
        table_version=table_version,
        **kwargs,
    )
    if not old_table_version:
        raise ValueError(
            f"Table version `{table_version}` does not exist for "
            f"table `{namespace}.{table_name}`."
        )
    new_table_version: TableVersion = Metafile.update_for(old_table_version)
    new_table_version.state = lifecycle_state or old_table_version.state
    # TODO(pdames): Use schema patch to check for backwards incompatible changes.
    #  By default, backwards incompatible changes should be pushed to a new
    #  table version unless the user explicitly forces the update to this
    #  table version (i.e., at the cost of potentially breaking consumers).
    update_schema = schema and not schema.equivalent_to(
        old_table_version.schema,
        True,
    )
    if update_schema and schema.id in [s.id for s in old_table_version.schemas]:
        raise ValueError(
            f"Schema ID `{schema.id}` already exists in "
            f"table version `{table_version}`."
        )
    new_table_version.schema = schema if update_schema else old_table_version.schema
    new_table_version.schemas = (
        old_table_version.schemas + [schema]
        if update_schema
        else old_table_version.schemas
    )
    new_table_version.description = (
        description if description is not None else old_table_version.description
    )
    new_table_version.properties = (
        properties if properties is not None else old_table_version.properties
    )
    new_table_version.partition_scheme = (
        partition_scheme or old_table_version.partition_scheme
    )
    # TODO(pdames): Check for backwards incompatible partition scheme changes.
    update_partition_scheme = partition_scheme and not partition_scheme.equivalent_to(
        old_table_version.partition_scheme,
        True,
    )
    if update_partition_scheme and partition_scheme.id in [
        ps.id for ps in old_table_version.partition_schemes
    ]:
        raise ValueError(
            f"Partition scheme ID `{partition_scheme.id}` already exists in "
            f"table version `{table_version}`."
        )
    new_table_version.partition_schemes = (
        old_table_version.partition_schemes + [partition_scheme]
        if update_partition_scheme
        else old_table_version.partition_schemes
    )
    # TODO(pdames): Check for backwards incompatible sort scheme changes.
    update_sort_scheme = sort_keys and not sort_keys.equivalent_to(
        old_table_version.sort_scheme,
        True,
    )
    if update_sort_scheme and sort_keys.id in [
        sk.id for sk in old_table_version.sort_schemes
    ]:
        raise ValueError(
            f"Sort scheme ID `{sort_keys.id}` already exists in "
            f"table version `{table_version}`."
        )
    new_table_version.sort_scheme = sort_keys or old_table_version.sort_scheme
    new_table_version.sort_schemes = (
        old_table_version.sort_schemes + [sort_keys]
        if update_sort_scheme
        else old_table_version.sort_schemes
    )
    old_table = get_table(
        *args,
        namespace=namespace,
        table_name=table_name,
        **kwargs,
    )
    txn_operations = []
    if (
        lifecycle_state == LifecycleState.ACTIVE
        and old_table_version.state != LifecycleState.ACTIVE
    ):
        _, old_version_number = (
            TableVersion.parse_table_version(
                old_table.latest_active_table_version,
            )
            if old_table.latest_active_table_version
            else (None, None)
        )
        _, new_version_number = TableVersion.parse_table_version(table_version)
        if old_version_number is None or old_version_number < new_version_number:
            # update the table's latest table version
            new_table: Table = Metafile.update_for(old_table)
            new_table.latest_active_table_version = table_version
            txn_operations.append(
                TransactionOperation.of(
                    operation_type=TransactionOperationType.UPDATE,
                    dest_metafile=new_table,
                    src_metafile=old_table,
                )
            )
    txn_operations.append(
        TransactionOperation.of(
            operation_type=TransactionOperationType.UPDATE,
            dest_metafile=new_table_version,
            src_metafile=old_table_version,
        ),
    )
    # TODO(pdames): Push changes down to non-deltacat streams via sync module.
    #   Also copy sort scheme changes down to deltacat child stream?
    if partition_scheme:
        old_stream = get_stream(
            *args,
            namespace=namespace,
            table_name=table_name,
            table_version=table_version,
            **kwargs,
        )
        new_stream: Stream = Metafile.update_for(old_stream)
        new_stream.partition_scheme = partition_scheme
        txn_operations.append(
            TransactionOperation.of(
                operation_type=TransactionOperationType.UPDATE,
                dest_metafile=new_stream,
                src_metafile=old_stream,
            )
        )
    transaction = Transaction.of(
        txn_type=TransactionType.ALTER,
        txn_operations=txn_operations,
    )
    catalog_properties = get_catalog_properties(**kwargs)
    transaction.commit(
        catalog_root_dir=catalog_properties.root,
        filesystem=catalog_properties.filesystem,
    )


def stage_stream(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    stream_format: StreamFormat = StreamFormat.DELTACAT,
    *args,
    **kwargs,
) -> Stream:
    """
    Stages a new delta stream for the given table version. Resolves to the
    latest active table version if no table version is given. Resolves to the
    DeltaCAT stream format if no stream format is given. If this stream
    will replace another stream with the same format and scheme, then it will
    have its previous stream ID set to the ID of the stream being replaced.
    Returns the staged stream. Raises an error if the table version does not
    exist.
    """
    # TODO(pdames): Support retrieving previously staged streams by ID.
    if not table_version:
        table_version = _resolve_latest_active_table_version_id(
            *args,
            namespace=namespace,
            table_name=table_name,
            **kwargs,
        )
    table_version_meta = get_table_version(
        *args,
        namespace=namespace,
        table_name=table_name,
        table_version=table_version,
        **kwargs,
    )
    locator = StreamLocator.at(
        namespace=namespace,
        table_name=table_name,
        table_version=table_version,
        stream_id=str(uuid.uuid4()),
        stream_format=stream_format or StreamFormat.DELTACAT,
    )
    stream = Stream.of(
        locator=locator,
        partition_scheme=table_version_meta.partition_scheme,
        state=CommitState.STAGED,
        previous_stream_id=None,
        watermark=None,
    )
    prev_stream = get_stream(
        *args,
        namespace=stream.namespace,
        table_name=stream.table_name,
        table_version=stream.table_version,
        stream_format=stream.stream_format,
        **kwargs,
    )
    if prev_stream:
        if prev_stream.stream_id == stream.stream_id:
            raise ValueError(
                f"Stream to stage has the same ID as existing stream: {prev_stream}."
            )
        stream.previous_stream_id = prev_stream.stream_id
    transaction = Transaction.of(
        txn_type=TransactionType.APPEND,
        txn_operations=[
            TransactionOperation.of(
                operation_type=TransactionOperationType.CREATE,
                dest_metafile=stream,
            )
        ],
    )
    catalog_properties = get_catalog_properties(**kwargs)
    transaction.commit(
        catalog_root_dir=catalog_properties.root,
        filesystem=catalog_properties.filesystem,
    )
    return stream


def commit_stream(
    stream: Stream,
    *args,
    **kwargs,
) -> Stream:
    """
    Registers a staged delta stream with a target table version, replacing any
    previous stream registered for the same table version. Returns the
    committed stream.
    """
    if not stream.stream_id:
        raise ValueError("Stream ID to commit must be set to a staged stream ID.")
    if not stream.table_version_locator:
        raise ValueError(
            "Stream to commit must have its table version locator "
            "set to the parent of its staged stream ID."
        )
    prev_staged_stream = get_stream_by_id(
        *args,
        table_version_locator=stream.table_version_locator,
        stream_id=stream.stream_id,
        **kwargs,
    )
    if not prev_staged_stream:
        raise ValueError(
            f"Stream at table version {stream.table_version_locator} with ID "
            f"{stream.stream_id} not found."
        )
    if prev_staged_stream.state != CommitState.STAGED:
        raise ValueError(
            f"Expected to find a `{CommitState.STAGED}` stream at table version "
            f"{stream.table_version_locator} with ID {stream.stream_id},"
            f"but found a `{prev_staged_stream.state}` partition."
        )
    if not prev_staged_stream:
        raise ValueError(
            f"Stream at table_version {stream.table_version_locator} with ID "
            f"{stream.stream_id} not found."
        )
    if prev_staged_stream.state != CommitState.STAGED:
        raise ValueError(
            f"Expected to find a `{CommitState.STAGED}` stream at table version "
            f"{stream.table_version_locator} with ID {stream.stream_id},"
            f"but found a `{prev_staged_stream.state}` stream."
        )
    stream: Stream = Metafile.update_for(prev_staged_stream)
    stream.state = CommitState.COMMITTED
    prev_committed_stream = get_stream(
        *args,
        namespace=stream.namespace,
        table_name=stream.table_name,
        table_version=stream.table_version,
        stream_format=stream.stream_format,
        **kwargs,
    )
    # the first transaction operation updates the staged stream commit state
    txn_type = TransactionType.ALTER
    txn_ops = [
        TransactionOperation.of(
            operation_type=TransactionOperationType.UPDATE,
            dest_metafile=stream,
            src_metafile=prev_staged_stream,
        )
    ]
    if prev_committed_stream:
        if prev_committed_stream.stream_id != stream.previous_stream_id:
            raise ValueError(
                f"Previous stream ID mismatch Expected "
                f"{stream.previous_stream_id} but found "
                f"{prev_committed_stream.stream_id}."
            )
        if prev_committed_stream.stream_id == stream.stream_id:
            raise ValueError(
                f"Stream to commit has the same ID as existing stream: {prev_committed_stream}."
            )
        # there's a previously committed stream, so update the transaction
        # type to overwrite the previously committed stream, and add another
        # transaction operation to replace it with the staged stream
        txn_type = TransactionType.OVERWRITE
        txn_ops.append(
            TransactionOperation.of(
                operation_type=TransactionOperationType.UPDATE,
                dest_metafile=stream,
                src_metafile=prev_committed_stream,
            )
        )
    transaction = Transaction.of(
        txn_type=txn_type,
        txn_operations=txn_ops,
    )
    catalog_properties = get_catalog_properties(**kwargs)
    transaction.commit(
        catalog_root_dir=catalog_properties.root,
        filesystem=catalog_properties.filesystem,
    )
    return stream


def delete_stream(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    stream_format: StreamFormat = StreamFormat.DELTACAT,
    *args,
    **kwargs,
) -> None:
    """
    Deletes the delta stream currently registered with the given table version.
    Resolves to the latest active table version if no table version is given.
    Resolves to the deltacat stream format if no stream format is given.
    Raises an error if the stream does not exist.
    """
    if not table_version:
        table_version = _resolve_latest_active_table_version_id(
            *args,
            namespace=namespace,
            table_name=table_name,
            **kwargs,
        )
    stream_to_delete = get_stream(
        *args,
        namespace=namespace,
        table_name=table_name,
        table_version=table_version,
        stream_format=stream_format,
        **kwargs,
    )
    if not stream_to_delete:
        raise ValueError(
            f"Stream to delete not found: {namespace}.{table_name}"
            f".{table_version}.{stream_format}."
        )
    else:
        stream_to_delete.state = CommitState.DEPRECATED
    transaction = Transaction.of(
        txn_type=TransactionType.DELETE,
        txn_operations=[
            TransactionOperation.of(
                operation_type=TransactionOperationType.DELETE,
                dest_metafile=stream_to_delete,
            )
        ],
    )
    catalog_properties = get_catalog_properties(**kwargs)
    transaction.commit(
        catalog_root_dir=catalog_properties.root,
        filesystem=catalog_properties.filesystem,
    )


def delete_table(
    namespace: str,
    name: str,
    purge: bool = False,
    *args,
    **kwargs,
) -> None:
    """
    Drops the given table and all its contents (table versions, streams, partitions,
    and deltas). If purge is True, also removes all data files associated with the table.
    Raises an error if the given table does not exist.

    TODO: Honor purge once garbage collection is implemented.
    """
    table: Optional[Table] = get_table(
        *args,
        namespace=namespace,
        table_name=name,
        **kwargs,
    )

    if not table:
        raise TableNotFoundError(f"Table `{namespace}.{name}` does not exist.")

    transaction = Transaction.of(
        txn_type=TransactionType.DELETE,
        txn_operations=TransactionOperationList.of(
            [
                TransactionOperation.of(
                    operation_type=TransactionOperationType.DELETE,
                    dest_metafile=table,
                )
            ]
        ),
    )

    catalog_properties = get_catalog_properties(**kwargs)
    transaction.commit(
        catalog_root_dir=catalog_properties.root,
        filesystem=catalog_properties.filesystem,
    )


def delete_namespace(
    namespace: str,
    purge: bool = False,
    *args,
    **kwargs,
) -> None:
    """
    Drops the given table namespace and all its contents. Raises an error if the
    given namespace does not exist.
    """
    namespace: Optional[Namespace] = get_namespace(
        *args,
        namespace=namespace,
        **kwargs,
    )

    if not namespace:
        raise ValueError(f"Namespace `{namespace}` does not exist.")

    transaction = Transaction.of(
        txn_type=TransactionType.DELETE,
        txn_operations=[
            TransactionOperation.of(
                operation_type=TransactionOperationType.DELETE,
                dest_metafile=namespace,
            )
        ],
    )
    catalog_properties = get_catalog_properties(**kwargs)
    transaction.commit(
        catalog_root_dir=catalog_properties.root,
        filesystem=catalog_properties.filesystem,
    )


def get_stream_by_id(
    table_version_locator: TableVersionLocator,
    stream_id: str,
    *args,
    **kwargs,
) -> Optional[Partition]:
    """
    Gets the stream for the given table version locator and stream ID.
    Returns None if the stream does not exist. Raises an error if the given
    table version locator does not exist.
    """
    locator = StreamLocator.of(
        table_version_locator=table_version_locator,
        stream_id=stream_id,
        stream_format=None,
    )
    return _latest(
        *args,
        metafile=Stream.of(locator=locator, partition_scheme=None),
        **kwargs,
    )


def get_stream(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    stream_format: StreamFormat = StreamFormat.DELTACAT,
    *args,
    **kwargs,
) -> Optional[Stream]:
    """
    Gets the most recently committed stream for the given table version.
    Resolves to the latest active table version if no table version is given.
    Resolves to the DeltaCAT stream format if no stream format is given.
    Returns None if the table version or stream format does not exist.
    """
    if not table_version:
        table_version = _resolve_latest_active_table_version_id(
            *args,
            namespace=namespace,
            table_name=table_name,
            fail_if_no_active_table_version=False,
            **kwargs,
        )
    locator = StreamLocator.at(
        namespace=namespace,
        table_name=table_name,
        table_version=table_version,
        stream_id=None,
        stream_format=stream_format,
    )
    return _latest(
        *args,
        metafile=Stream.of(
            locator=locator,
            partition_scheme=None,
            state=CommitState.COMMITTED,
        ),
        **kwargs,
    )


def stream_exists(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    stream_format: StreamFormat = StreamFormat.DELTACAT,
    *args,
    **kwargs,
) -> Optional[Stream]:
    """
    Returns True if the given Stream exists, False if not.
    Resolves to the latest active table version if no table version is given.
    Resolves to the DeltaCAT stream format if no stream format is given.
    Returns None if the table version or stream format does not exist.
    """
    if not table_version:
        table_version = _resolve_latest_active_table_version_id(
            *args,
            namespace=namespace,
            table_name=table_name,
            fail_if_no_active_table_version=False,
            **kwargs,
        )
    locator = StreamLocator.at(
        namespace=namespace,
        table_name=table_name,
        table_version=table_version,
        stream_id=None,
        stream_format=stream_format,
    )
    return _exists(
        *args,
        metafile=Stream.of(
            locator=locator,
            partition_scheme=None,
            state=CommitState.COMMITTED,
        ),
        **kwargs,
    )


def stage_partition(
    stream: Stream,
    partition_values: Optional[PartitionValues] = None,
    partition_scheme_id: Optional[str] = None,
    *args,
    **kwargs,
) -> Partition:
    """
    Stages a new partition for the given stream and partition values. Returns
    the staged partition. If this partition will replace another partition
    with the same partition values and scheme, then it will have its previous
    partition ID set to the ID of the partition being replaced. Partition values
    should not be specified for unpartitioned tables.

    The partition_values must represent the results of transforms in a partition
    spec specified in the stream.
    """
    # TODO(pdames): Cache last retrieved metafile revisions in memory to resolve
    #   potentially high cost of staging many partitions.
    table_version = get_table_version(
        *args,
        namespace=stream.namespace,
        table_name=stream.table_name,
        table_version=stream.table_version,
        **kwargs,
    )
    if not table_version:
        raise ValueError(
            f"Table version not found: {stream.namespace}.{stream.table_name}."
            f"{stream.table_version}."
        )
    if not table_version.partition_schemes or partition_scheme_id not in [
        ps.id for ps in table_version.partition_schemes
    ]:
        raise ValueError(
            f"Invalid partition scheme ID `{partition_scheme_id}` (not found "
            f"in parent table version `{stream.namespace}.{stream.table_name}"
            f".{table_version.table_version}` partition scheme IDs)."
        )
    if stream.partition_scheme.id not in table_version.partition_schemes:
        # this should never happen, but just in case
        raise ValueError(
            f"Invalid stream partition scheme ID `{stream.partition_scheme.id}`"
            f"in parent table version `{stream.namespace}.{stream.table_name}"
            f".{table_version.table_version}` partition scheme IDs)."
        )
    locator = PartitionLocator.of(
        stream_locator=stream.locator,
        partition_values=partition_values,
        partition_id=str(uuid.uuid4()),
    )
    partition = Partition.of(
        locator=locator,
        schema=table_version.schema,
        content_types=table_version.content_types,
        state=CommitState.STAGED,
        previous_stream_position=None,
        partition_values=partition_values,
        previous_partition_id=None,
        stream_position=None,
        partition_scheme_id=partition_scheme_id,
    )
    prev_partition = get_partition(
        *args,
        stream_locator=stream.locator,
        partition_values=partition_values,
        partition_scheme_id=partition_scheme_id,
        **kwargs,
    )
    if prev_partition:
        if prev_partition.partition_id == partition.partition_id:
            raise ValueError(
                f"Partition to stage has the same ID as existing partition: {prev_partition}."
            )
        partition.previous_partition_id = prev_partition.partition_id
    transaction = Transaction.of(
        txn_type=TransactionType.APPEND,
        txn_operations=[
            TransactionOperation.of(
                operation_type=TransactionOperationType.CREATE,
                dest_metafile=partition,
            )
        ],
    )
    catalog_properties = get_catalog_properties(**kwargs)
    transaction.commit(
        catalog_root_dir=catalog_properties.root,
        filesystem=catalog_properties.filesystem,
    )
    return partition


def commit_partition(
    partition: Partition,
    previous_partition: Optional[Partition] = None,
    *args,
    **kwargs,
) -> Partition:
    """
    Commits the staged partition to its associated table version stream,
    replacing any previous partition registered for the same stream and
    partition values.

    If previous partition is given then it will be replaced with its deltas
    prepended to the new partition being committed. Otherwise the latest
    committed partition with the same keys and partition scheme ID will be
    retrieved.

    Returns the registered partition. If the partition's
    previous delta stream position is specified, then the commit will
    be rejected if it does not match the actual previous stream position of
    the partition being replaced. If the partition's previous partition ID is
    specified, then the commit will be rejected if it does not match the actual
    ID of the partition being replaced.
    """
    if previous_partition:
        raise NotImplementedError(
            f"delta prepending from previous partition {previous_partition} "
            f"is not yet implemented"
        )
    if not partition.partition_id:
        raise ValueError("Partition ID to commit must be set to a staged partition ID.")
    if not partition.stream_locator:
        raise ValueError(
            "Partition to commit must have its stream locator "
            "set to the parent of its staged partition ID."
        )
    prev_staged_partition = get_partition_by_id(
        *args,
        stream_locator=partition.stream_locator,
        partition_id=partition.partition_id,
        **kwargs,
    )
    if not prev_staged_partition:
        raise ValueError(
            f"Partition at stream {partition.stream_locator} with ID "
            f"{partition.partition_id} not found."
        )
    if prev_staged_partition.state != CommitState.STAGED:
        raise ValueError(
            f"Expected to find a `{CommitState.STAGED}` partition at stream "
            f"{partition.stream_locator} with ID {partition.partition_id},"
            f"but found a `{prev_staged_partition.state}` partition."
        )
    partition: Partition = Metafile.update_for(prev_staged_partition)
    partition.state = CommitState.COMMITTED
    prev_committed_partition = get_partition(
        *args,
        stream_locator=partition.stream_locator,
        partition_value=partition.partition_values,
        partition_scheme_id=partition.partition_scheme_id,
        **kwargs,
    )
    # the first transaction operation updates the staged partition commit state
    txn_type = TransactionType.ALTER
    txn_ops = [
        TransactionOperation.of(
            operation_type=TransactionOperationType.UPDATE,
            dest_metafile=partition,
            src_metafile=prev_staged_partition,
        )
    ]
    if prev_committed_partition:
        if prev_committed_partition.partition_id != partition.previous_partition_id:
            raise ValueError(
                f"Previous partition ID mismatch Expected "
                f"{partition.previous_partition_id} but found "
                f"{prev_committed_partition.partition_id}."
            )
        # TODO(pdames): Add previous partition stream position validation.
        if prev_committed_partition.partition_id == partition.partition_id:
            raise ValueError(
                f"Partition to commit has the same ID as existing partition: "
                f"{prev_committed_partition}."
            )
        # there's a previously committed partition, so update the transaction
        # type to overwrite the previously committed partition, and add another
        # transaction operation to replace it with the staged partition
        txn_type = TransactionType.OVERWRITE
        txn_ops.append(
            TransactionOperation.of(
                operation_type=TransactionOperationType.UPDATE,
                dest_metafile=partition,
                src_metafile=prev_committed_partition,
            )
        )
    transaction = Transaction.of(
        txn_type=txn_type,
        txn_operations=txn_ops,
    )
    catalog_properties = get_catalog_properties(**kwargs)
    transaction.commit(
        catalog_root_dir=catalog_properties.root,
        filesystem=catalog_properties.filesystem,
    )
    return partition


def delete_partition(
    stream_locator: StreamLocator,
    partition_values: Optional[PartitionValues] = None,
    partition_scheme_id: Optional[str] = None,
    *args,
    **kwargs,
) -> None:
    """
    Deletes the given partition from the specified stream. Partition
    values should not be specified for unpartitioned tables. Raises an error
    if the partition does not exist.
    """
    partition_to_delete = get_partition(
        *args,
        stream_locator=stream_locator,
        partition_values=partition_values,
        partition_scheme_id=partition_scheme_id,
        **kwargs,
    )
    if not partition_to_delete:
        raise ValueError(
            f"Partition with values {partition_values} and scheme "
            f"{partition_scheme_id} not found in stream: {stream_locator}"
        )
    else:
        partition_to_delete.state = CommitState.DEPRECATED
    transaction = Transaction.of(
        txn_type=TransactionType.DELETE,
        txn_operations=[
            TransactionOperation.of(
                operation_type=TransactionOperationType.DELETE,
                src_metafile=partition_to_delete,
            )
        ],
    )
    catalog_properties = get_catalog_properties(**kwargs)
    transaction.commit(
        catalog_root_dir=catalog_properties.root,
        filesystem=catalog_properties.filesystem,
    )


def get_partition_by_id(
    stream_locator: StreamLocator,
    partition_id: str,
    *args,
    **kwargs,
) -> Optional[Partition]:
    """
    Gets the partition for the given stream locator and partition ID.
    Returns None if the partition does not exist. Raises an error if the
    given stream locator does not exist.
    """
    locator = PartitionLocator.of(
        stream_locator=stream_locator,
        partition_values=None,
        partition_id=partition_id,
    )
    return _latest(
        *args,
        metafile=Partition.of(
            locator=locator,
            schema=None,
            content_types=None,
        ),
        **kwargs,
    )


def get_partition(
    stream_locator: StreamLocator,
    partition_values: Optional[PartitionValues] = None,
    partition_scheme_id: Optional[str] = None,
    *args,
    **kwargs,
) -> Optional[Partition]:
    """
    Gets the most recently committed partition for the given stream locator and
    partition key values. Returns None if no partition has been committed for
    the given table version and/or partition key values. Partition values
    should not be specified for unpartitioned tables. Partition scheme ID
    resolves to the table version's current partition scheme by default.
    Raises an error if the given stream locator does not exist.
    """
    locator = PartitionLocator.of(
        stream_locator=stream_locator,
        partition_values=partition_values,
        partition_id=None,
    )
    if not partition_scheme_id:
        # resolve latest partition scheme from the current
        # revision of its `deltacat` stream
        stream = get_stream(
            *args,
            namespace=stream_locator.namespace,
            table_name=stream_locator.table_name,
            table_version=stream_locator.table_version,
            **kwargs,
        )
        if not stream:
            raise ValueError(f"Stream {stream_locator} not found.")
        partition_scheme_id = stream.partition_scheme.id
    return _latest(
        *args,
        metafile=Partition.of(
            locator=locator,
            schema=None,
            content_types=None,
            state=CommitState.COMMITTED,
            partition_scheme_id=partition_scheme_id,
        ),
        **kwargs,
    )


def stage_delta(
    data: Union[LocalTable, LocalDataset, DistributedDataset, Manifest],
    partition: Partition,
    delta_type: DeltaType = DeltaType.UPSERT,
    max_records_per_entry: Optional[int] = None,
    author: Optional[ManifestAuthor] = None,
    properties: Optional[DeltaProperties] = None,
    s3_table_writer_kwargs: Optional[Dict[str, Any]] = None,
    content_type: ContentType = ContentType.PARQUET,
    entry_params: Optional[EntryParams] = None,
    *args,
    **kwargs,
) -> Delta:
    """
    Writes the given table to 1 or more S3 files. Returns an unregistered
    delta whose manifest entries point to the uploaded files. Applies any
    schema consistency policies configured for the parent table version.

    The partition spec will be used to split the input table into
    multiple files. Optionally, partition_values can be provided to avoid
    this method to recompute partition_values from the provided data.

    Raises an error if the provided data does not conform to a unique ordered
    list of partition_values
    """
    raise NotImplementedError("stage_delta not implemented")


def commit_delta(delta: Delta, *args, **kwargs) -> Delta:
    """
    Registers a new delta with its associated target table version and
    partition. Returns the registered delta. If the delta's previous stream
    position is specified, then the commit will be rejected if it does not match
    the target partition's actual previous stream position. If the delta's
    stream position is specified, it must be greater than the latest stream
    position in the target partition.
    """
    raise NotImplementedError("commit_delta not implemented")


def get_namespace(namespace: str, *args, **kwargs) -> Optional[Namespace]:
    """
    Gets table namespace metadata for the specified table namespace. Returns
    None if the given namespace does not exist.
    """
    return _latest(
        *args,
        metafile=Namespace.of(NamespaceLocator.of(namespace)),
        **kwargs,
    )


def namespace_exists(namespace: str, *args, **kwargs) -> bool:
    """
    Returns True if the given table namespace exists, False if not.
    """
    return _exists(
        *args,
        metafile=Namespace.of(NamespaceLocator.of(namespace)),
        **kwargs,
    )


def get_table(namespace: str, table_name: str, *args, **kwargs) -> Optional[Table]:
    """
    Gets table metadata for the specified table. Returns None if the given
    table does not exist.
    """
    locator = TableLocator.at(namespace=namespace, table_name=table_name)
    return _latest(
        *args,
        metafile=Table.of(locator=locator),
        **kwargs,
    )


def table_exists(namespace: str, table_name: str, *args, **kwargs) -> bool:
    """
    Returns True if the given table exists, False if not.
    """
    locator = TableLocator.at(namespace=namespace, table_name=table_name)
    return _exists(
        *args,
        metafile=Table.of(locator=locator),
        **kwargs,
    )


def get_table_version(
    namespace: str,
    table_name: str,
    table_version: str,
    *args,
    **kwargs,
) -> Optional[TableVersion]:
    """
    Gets table version metadata for the specified table version. Returns None
    if the given table version does not exist.
    """
    locator = TableVersionLocator.at(
        namespace=namespace,
        table_name=table_name,
        table_version=table_version,
    )
    table_version = TableVersion.of(
        locator=locator,
        schema=None,
    )
    return _latest(
        *args,
        metafile=table_version,
        **kwargs,
    )


def get_latest_table_version(
    namespace: str, table_name: str, *args, **kwargs
) -> Optional[TableVersion]:
    """
    Gets table version metadata for the latest version of the specified table.
    Returns None if no table version exists for the given table. Raises
    an error if the given table doesn't exist.
    """
    table_version_id = _resolve_latest_table_version_id(
        *args,
        namespace=namespace,
        table_name=table_name,
        fail_if_no_active_table_version=False,
        **kwargs,
    )

    return (
        get_table_version(
            *args,
            namespace=namespace,
            table_name=table_name,
            table_version=table_version_id,
            **kwargs,
        )
        if table_version_id
        else None
    )


def get_latest_active_table_version(
    namespace: str, table_name: str, *args, **kwargs
) -> Optional[TableVersion]:
    """
    Gets table version metadata for the latest active version of the specified
    table. Returns None if no active table version exists for the given table.
    Raises an error if the given table doesn't exist.
    """
    table_version_id = _resolve_latest_active_table_version_id(
        *args,
        namespace=namespace,
        table_name=table_name,
        fail_if_no_active_table_version=False,
        **kwargs,
    )
    return (
        get_table_version(
            *args,
            namespace=namespace,
            table_name=table_name,
            table_version=table_version_id,
            **kwargs,
        )
        if table_version_id
        else None
    )


def get_table_version_column_names(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    *args,
    **kwargs,
) -> Optional[List[str]]:
    """
    Gets a list of column names for the specified table version, or for the
    latest active table version if none is specified. The index of each
    column name returned represents its ordinal position in a delimited text
    file or other row-oriented content type files appended to the table.
    Returns None for schemaless tables. Raises an error if the table version
    does not exist.
    """
    schema = get_table_version_schema(
        namespace=namespace,
        table_name=table_name,
        table_version=table_version,
    )
    return schema.arrow.names if schema else None


def get_table_version_schema(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    *args,
    **kwargs,
) -> Optional[Schema]:
    """
    Gets the schema for the specified table version, or for the latest active
    table version if none is specified. Returns None if the table version is
    schemaless. Raises an error if the table version does not exist.
    """
    table_version = (
        get_table_version(
            *args,
            namespace=namespace,
            table_name=table_name,
            table_version=table_version,
            **kwargs,
        )
        if table_version
        else get_latest_active_table_version(
            *args,
            namespace=namespace,
            table_name=table_name,
            **kwargs,
        )
    )
    return table_version.schema


def table_version_exists(
    namespace: str,
    table_name: str,
    table_version: str,
    *args,
    **kwargs,
) -> bool:
    """
    Returns True if the given table version exists, False if not.
    """
    locator = TableVersionLocator.at(
        namespace=namespace,
        table_name=table_name,
        table_version=table_version,
    )
    table_version = TableVersion.of(
        locator=locator,
        schema=None,
    )
    return _exists(
        *args,
        metafile=table_version,
        **kwargs,
    )


def can_categorize(e: BaseException, *args, **kwargs) -> bool:
    """
    Return whether input error is from storage implementation layer.
    """
    raise NotImplementedError


def raise_categorized_error(e: BaseException, *args, **kwargs):
    """
    Raise and handle storage implementation layer specific errors.
    """
    raise NotImplementedError



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

from typing import Optional, Any, List, Tuple, Dict

from pyarrow.compute import SortOptions

from deltacat.storage.model.types import (
    SortOrder,
    NullOrder,
)
from deltacat.storage.model.schema import FieldLocator
from deltacat.storage.model.transform import Transform


class SortKey(tuple):
    @staticmethod
    def of(
        key: Optional[List[FieldLocator]],
        sort_order: SortOrder = SortOrder.ASCENDING,
        null_order: NullOrder = NullOrder.AT_END,
        transform: Optional[Transform] = None,
        native_object: Optional[Any] = None,
    ) -> SortKey:
        """
        Create a sort key from a field name to use as the sort key, and
        the sort order for this key. If no sort order is specified, then the
        data will be sorted in ascending order by default.
        """
        return SortKey(
            (
                key,
                sort_order.value if isinstance(sort_order, SortOrder) else sort_order,
                null_order.value if isinstance(null_order, NullOrder) else null_order,
                transform,
                native_object,
            )
        )

    def equivalent_to(
        self,
        other: SortKey,
    ):
        if other is None:
            return False
        if not isinstance(other, tuple):
            return False
        if not isinstance(other, SortKey):
            other = SortKey(other)
        return (
            self.key == other.key
            and self.transform == other.transform
            and self.sort_order == other.sort_order
            and self.null_order == other.null_order
        )

    @property
    def key(self) -> Optional[List[FieldLocator]]:
        return self[0]

    @property
    def sort_order(self) -> SortOrder:
        return SortOrder(self[1])

    @property
    def null_order(self) -> NullOrder:
        return NullOrder(self[2])

    @property
    def transform(self) -> Optional[Transform]:
        val: Dict[str, Any] = (
            Transform(self[3]) if len(self) >= 4 and self[3] is not None else None
        )
        return val

    @property
    def arrow(self) -> List[Tuple[str, str]]:
        # TODO(pdames): Convert unsupported field locators to arrow field names,
        #   and transforms/multi-key-sorts to pyarrow compute expressions. Add
        #   null order via SortOptions when supported per field by Arrow.
        return (
            [(field_locator, self[1]) for field_locator in self[0]] if self[0] else []
        )

    @property
    def native_object(self) -> Optional[Any]:
        return self[4] if len(self) >= 5 else None


class SortKeyList(List[SortKey]):
    @staticmethod
    def of(items: List[SortKey]) -> SortKeyList:
        typed_items = SortKeyList()
        for item in items:
            if item is not None and not isinstance(item, SortKey):
                item = SortKey(item)
            typed_items.append(item)
        return typed_items

    def __getitem__(self, item):
        val = super().__getitem__(item)
        if val is not None and not isinstance(val, SortKey):
            self[item] = val = SortKey(val)
        return val


class SortScheme(dict):
    @staticmethod
    def of(
        keys: Optional[SortKeyList],
        name: Optional[str] = None,
        scheme_id: Optional[str] = None,
        native_object: Optional[Any] = None,
    ) -> SortScheme:
        return SortScheme(
            {
                "keys": keys,
                "name": name,
                "id": scheme_id,
                "nativeObject": native_object,
            }
        )

    def equivalent_to(
        self,
        other: SortScheme,
        check_identifiers: bool = False,
    ) -> bool:
        if other is None:
            return False
        if not isinstance(other, dict):
            return False
        if not isinstance(other, SortScheme):
            other = SortScheme(other)
        for i in range(len(self.keys)):
            if not self.keys[i].equivalent_to(other.keys[i]):
                return False
        return not check_identifiers or (
            self.name == other.name and self.id == other.id
        )

    @property
    def keys(self) -> Optional[SortKeyList]:
        val: List[SortKey] = self.get("keys")
        if val is not None and not isinstance(val, SortKeyList):
            self["keys"] = val = SortKeyList.of(val)
        return val

    @property
    def name(self) -> Optional[str]:
        return self.get("name")

    @property
    def id(self) -> Optional[str]:
        return self.get("id")

    @property
    def arrow(self) -> SortOptions:
        # TODO(pdames): Remove homogenous null ordering when supported by Arrow.
        if self.keys:
            if len(set([key.null_order for key in self.keys])) == 1:
                return SortOptions(
                    sort_keys=[pa_key for k in self.keys for pa_key in k.arrow],
                    null_placement=self.keys[0].null_order.value,
                )
            else:
                err_msg = "All arrow sort keys must use the same null order."
                raise ValueError(err_msg)
        return SortOptions()

    @property
    def native_object(self) -> Optional[Any]:
        return self.get("nativeObject")


class SortSchemeList(List[SortScheme]):
    @staticmethod
    def of(items: List[SortScheme]) -> SortSchemeList:
        typed_items = SortSchemeList()
        for item in items:
            if item is not None and not isinstance(item, SortScheme):
                item = SortScheme(item)
            typed_items.append(item)
        return typed_items

    def __getitem__(self, item):
        val = super().__getitem__(item)
        if val is not None and not isinstance(val, SortScheme):
            self[item] = val = SortScheme(val)
        return val



from deltacat.catalog.model.properties import (  # noqa: F401
    CatalogProperties,
    get_catalog_properties,
)
from deltacat.catalog.model.catalog import Catalog, Catalogs  # noqa: F401
from deltacat.catalog.main import impl as DeltacatCatalog

__all__ = [
    "CatalogProperties",
    "get_catalog_properties",
    "Catalog",
    "Catalogs",
    "DeltacatCatalog",
]



import shutil
import tempfile

import pytest
import copy
import pyarrow as pa

from deltacat import PartitionKey, PartitionScheme
from deltacat.storage import (
    metastore,
    CommitState,
    IdentityTransform,
    LifecycleState,
    Metafile,
    Namespace,
    NamespaceLocator,
    TableVersion,
    TableVersionLocator,
    Schema,
    SortKey,
    SortScheme,
    Stream,
    StreamFormat,
)
from deltacat.tests.test_utils.storage import (
    create_test_namespace,
    create_test_table,
    create_test_table_version,
)
from deltacat.catalog import CatalogProperties

from deltacat.storage.main.impl import DEFAULT_TABLE_VERSION


class TestNamespace:
    @classmethod
    def setup_method(cls):

        cls.tmpdir = tempfile.mkdtemp()
        cls.catalog = CatalogProperties(root=cls.tmpdir)
        cls.namespace1 = metastore.create_namespace(
            namespace="namespace1",
            catalog=cls.catalog,
        )
        cls.namespace2 = metastore.create_namespace(
            namespace="namespace2",
            catalog=cls.catalog,
        )

    @classmethod
    def teardown_method(cls):
        shutil.rmtree(cls.tmpdir)

    def test_list_namespaces(self):
        # expect the namespace returned to match the input namespace to create
        namespace_locator = NamespaceLocator.of(namespace="namespace1")
        expected_namespace = Namespace.of(locator=namespace_locator)
        assert expected_namespace.equivalent_to(self.namespace1)

        # expect the namespace to exist
        assert metastore.namespace_exists(
            namespace="namespace1",
            catalog=self.catalog,
        )

        # expect the namespace to also be returned when listing namespaces
        list_result = metastore.list_namespaces(catalog=self.catalog)
        namespaces_by_name = {n.locator.namespace: n for n in list_result.all_items()}
        assert len(namespaces_by_name.items()) == 2
        assert namespaces_by_name["namespace1"].equivalent_to(self.namespace1)
        assert namespaces_by_name["namespace2"].equivalent_to(self.namespace2)

    def test_get_namespace(self):
        # expect the namespace to also be returned when explicitly retrieved
        read_namespace = metastore.get_namespace(
            namespace="namespace1",
            catalog=self.catalog,
        )
        assert read_namespace and read_namespace.equivalent_to(self.namespace1)

    def test_namespace_exists_existing(self):
        assert metastore.namespace_exists(
            "namespace1",
            catalog=self.catalog,
        )

    def test_namespace_not_exists(self):
        assert not metastore.namespace_exists(
            "foobar",
            catalog=self.catalog,
        )


class TestTable:
    @classmethod
    def setup_method(cls):
        cls.tmpdir = tempfile.mkdtemp()
        cls.catalog = CatalogProperties(root=cls.tmpdir)
        # Create a namespace to hold our tables
        cls.test_namespace = create_test_namespace()
        cls.namespace_obj = metastore.create_namespace(
            namespace=cls.test_namespace.namespace,
            catalog=cls.catalog,
        )
        cls.test_table1 = create_test_table()
        cls.test_table1.latest_table_version = "v.1"
        cls.test_table2 = create_test_table()
        cls.test_table2.locator.table_name = "table2"
        cls.test_table2.latest_table_version = "v.1"
        # Create two table versions (their parent tables will be auto-created)
        cls.table1, cls.tv1, cls.stream1 = metastore.create_table_version(
            namespace=cls.test_table1.namespace,
            table_name=cls.test_table1.table_name,
            table_version=cls.test_table1.latest_table_version,
            table_description=cls.test_table1.description,
            table_properties=cls.test_table1.properties,
            catalog=cls.catalog,
        )
        cls.table2, cls.tv2, cls.stream2 = metastore.create_table_version(
            namespace=cls.test_table2.namespace,
            table_name=cls.test_table2.table_name,
            table_version=cls.test_table2.latest_table_version,
            table_description=cls.test_table2.description,
            table_properties=cls.test_table2.properties,
            catalog=cls.catalog,
        )

    @classmethod
    def teardown_method(cls):
        shutil.rmtree(cls.tmpdir)

    def test_list_tables(self):
        # list the tables under our namespace
        list_result = metastore.list_tables(
            namespace=self.test_namespace.namespace,
            catalog=self.catalog,
        )
        all_tables = list_result.all_items()

        # we expect 2 distinct tables
        for table in all_tables:
            if table.table_name == self.test_table1.table_name:
                assert table.equivalent_to(self.test_table1)
            else:
                assert table.equivalent_to(self.test_table2)

    def test_get_table(self):
        # test we can retrieve table1 by name
        tbl = metastore.get_table(
            namespace=self.test_namespace.namespace,
            table_name=self.test_table1.table_name,
            catalog=self.catalog,
        )
        assert tbl is not None
        assert tbl.equivalent_to(self.test_table1)

    def test_table_exists_existing(self):
        # table1 should exist
        assert metastore.table_exists(
            namespace=self.test_namespace.namespace,
            table_name=self.test_table1.table_name,
            catalog=self.catalog,
        )

    def test_table_not_exists(self):
        assert not metastore.table_exists(
            namespace=self.test_namespace.namespace,
            table_name="no_such_table",
            catalog=self.catalog,
        )


class TestTableVersion:
    @classmethod
    def setup_method(cls):
        cls.tmpdir = tempfile.mkdtemp()
        cls.catalog = CatalogProperties(root=cls.tmpdir)

        # create the namespace that we'll attach the base table to
        cls.namespace = create_test_namespace()
        # create the base table that we'll attach table versions to
        cls.table = create_test_table()
        # create the first table version to attach to the base table
        cls.table_version = create_test_table_version()
        # create the second table version to attach to the base table
        cls.table_version2 = create_test_table_version()
        cls.table_version2.previous_table_version = cls.table_version.table_version
        cls.table_version2.locator.table_version = "v.2"

        # create a namespace and single table
        cls.namespace_obj = metastore.create_namespace(
            namespace=cls.namespace.namespace,
            catalog=cls.catalog,
        )

        # create a "base" table with single table version attached
        cls.table1, cls.tv1, cls.stream1 = metastore.create_table_version(
            namespace=cls.table.namespace,
            table_name=cls.table.table_name,
            table_version=cls.table_version.table_version,
            schema=cls.table_version.schema,
            partition_scheme=cls.table_version.partition_scheme,
            sort_keys=cls.table_version.sort_scheme,
            table_version_description=cls.table_version.description,
            table_version_properties=cls.table_version.properties,
            table_description=cls.table.description,
            table_properties=cls.table.properties,
            supported_content_types=cls.table_version.content_types,
            catalog=cls.catalog,
        )
        # now attach a second table version to the same base table
        cls.table2, cls.tv2, cls.stream2 = metastore.create_table_version(
            namespace=cls.table.namespace,
            table_name=cls.table.table_name,
            table_version=cls.table_version2.table_version,
            schema=cls.table_version2.schema,
            partition_scheme=cls.table_version2.partition_scheme,
            sort_keys=cls.table_version2.sort_scheme,
            table_version_description=cls.table_version2.description,
            table_version_properties=cls.table_version2.properties,
            table_description=cls.table.description,
            table_properties=cls.table.properties,
            supported_content_types=cls.table_version2.content_types,
            catalog=cls.catalog,
        )
        cls.table.latest_table_version = cls.table_version2.table_version

    @classmethod
    def teardown_method(cls):
        shutil.rmtree(cls.tmpdir)

    def test_create_bad_next_table_version(self):
        # given that the latest ordinal table version is 2
        table_version = create_test_table_version()
        # when we try to create ordinal table version 1 again
        # expect an error to be raised (ordinal version 3 expected)
        with pytest.raises(ValueError):
            metastore.create_table_version(
                namespace=self.table.namespace,
                table_name=self.table.table_name,
                table_version=table_version.table_version,
                schema=self.table_version.schema,
                partition_scheme=table_version.partition_scheme,
                sort_keys=self.table_version.sort_scheme,
                table_version_description=table_version.description,
                table_version_properties=table_version.properties,
                table_description=self.table.description,
                table_properties=self.table.properties,
                supported_content_types=table_version.content_types,
                catalog=self.catalog,
            )

    def test_create_next_table_version(self):
        # given that our test table's latest ordinal table version is 2
        table_version = create_test_table_version()
        table_version.locator.table_version = TableVersion.next_version(
            self.table_version2.table_version
        )
        # when we try to create the next ordinal table version (3)
        metastore.create_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=table_version.table_version,
            schema=self.table_version.schema,
            partition_scheme=table_version.partition_scheme,
            sort_keys=self.table_version.sort_scheme,
            table_version_description=table_version.description,
            table_version_properties=table_version.properties,
            table_description=self.table.description,
            table_properties=self.table.properties,
            supported_content_types=table_version.content_types,
            catalog=self.catalog,
        )
        # expect ordinal table version 3 to be successfully created
        table_version3 = metastore.get_latest_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            catalog=self.catalog,
        )
        table_version.previous_table_version = self.table_version2.table_version
        assert table_version3.equivalent_to(table_version)

    def test_create_first_table_version_default_id_assignment(self):
        # given a new first table version created without a table version ID
        metastore.create_table_version(
            namespace=self.table.namespace,
            table_name="test_table_2",
            schema=self.table_version.schema,
            partition_scheme=self.table_version.partition_scheme,
            sort_keys=self.table_version.sort_scheme,
            table_version_description=self.table_version.description,
            table_version_properties=self.table_version.properties,
            table_description=self.table.description,
            table_properties=self.table.properties,
            supported_content_types=self.table_version.content_types,
            catalog=self.catalog,
        )
        # when we retrieve this table version
        table_version = metastore.get_latest_table_version(
            namespace=self.table.namespace,
            table_name="test_table_2",
            catalog=self.catalog,
        )
        # expect it to have the correct default table version ID assigned
        table_version.previous_table_version = self.table_version2.table_version
        table_version_default_id: TableVersion = Metafile.update_for(self.table_version)
        table_version_default_id.locator.table_version = DEFAULT_TABLE_VERSION
        assert table_version.equivalent_to(table_version)

    def test_list_table_versions(self):
        # given 2 previously created table versions in the same table
        list_result = metastore.list_table_versions(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            catalog=self.catalog,
        )
        # when we list all table versions
        # expect the table versions fetched to be equivalent to those created
        tvs = list_result.all_items()
        for tv in tvs:
            if tv.id == self.table_version.id:
                assert tv.equivalent_to(self.table_version)
            elif tv.id == self.table_version2.id:
                assert tv.equivalent_to(self.table_version2)

    def test_list_table_versions_bad_parent_locator(self):
        kwargs = {
            "namespace": self.table.namespace,
            "table_name": self.table.table_name,
        }
        for key in kwargs.keys():
            kwargs_copy = copy.copy(kwargs)
            # given a bad table version parent locator
            kwargs_copy[key] = "i_dont_exist"
            # when we list table versions
            # expect an error to be raised
            with pytest.raises(ValueError):
                metastore.list_table_versions(
                    catalog=self.catalog,
                    **kwargs_copy,
                )

    def test_get_latest_table_version(self):
        # given two previously created table versions in the same table
        # when we get the latest table version
        tv = metastore.get_latest_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            catalog=self.catalog,
        )
        # expect it to be equivalent ot the last created table version
        assert tv.equivalent_to(self.table_version2)

    def test_get_latest_table_version_bad_parent_locator(self):
        kwargs = {
            "namespace": self.table.namespace,
            "table_name": self.table.table_name,
        }
        for key in kwargs.keys():
            kwargs_copy = copy.copy(kwargs)
            # given a bad table version parent locator
            kwargs_copy[key] = "i_dont_exist"
            # when we get the latest table version
            # expect an error to be raised
            with pytest.raises(ValueError):
                metastore.get_latest_table_version(
                    catalog=self.catalog,
                    **kwargs_copy,
                )

    def test_update_table_version_schema_add_named_subschema(self):
        # given an update to the schema of table version 1
        old_schema = self.table_version.schema
        new_pyarrow_schema = pa.schema(
            [
                ("col_1", pa.int64()),
                ("col_2", pa.float64()),
                ("col_3", pa.string()),
            ]
        )
        new_schema = old_schema.add_subschema(
            name="test",
            schema=new_pyarrow_schema,
        )
        metastore.update_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            schema=new_schema,
            catalog=self.catalog,
        )
        # when we get the new schema of table version 1
        actual_schema = metastore.get_table_version_schema(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            catalog=self.catalog,
        )
        # expect it to be equivalent to the expected schema
        assert actual_schema.equivalent_to(new_schema)
        assert not actual_schema.equivalent_to(old_schema)
        # expect the table version to have two schemas in its evolution history
        tv = metastore.get_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            catalog=self.catalog,
        )
        assert len(tv.schemas) == 2
        assert tv.schemas[0].equivalent_to(old_schema)
        assert tv.schemas[1].equivalent_to(new_schema)
        # expect ONLY the schema to be updated
        expected_tv = Metafile.update_for(self.table_version)
        expected_tv.schema = tv.schema
        expected_tv.schemas = [old_schema, tv.schema]
        assert tv.equivalent_to(expected_tv)

    def test_update_table_version_schema_same_schema_id_fails(self):
        # given an update to the schema of table version 1 w/ the same schema ID
        old_schema = self.table_version.schema
        new_schema = Schema.of(
            schema=pa.schema(
                [
                    ("col_1", pa.int64()),
                    ("col_2", pa.float64()),
                    ("col_3", pa.string()),
                ]
            ),
            schema_id=old_schema.id,
        )
        # when we try to update the schema
        # expect an error to be raised
        with pytest.raises(ValueError):
            metastore.update_table_version(
                namespace=self.table.namespace,
                table_name=self.table.table_name,
                table_version=self.table_version.table_version,
                schema=new_schema,
                catalog=self.catalog,
            )

    def test_update_table_version_schema_equivalent_schema_noop(self):
        # given a noop update to the schema of table version 1
        old_schema = self.table_version.schema
        new_schema = Schema.of(
            schema=old_schema.arrow,
        )
        metastore.update_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            schema=new_schema,
            catalog=self.catalog,
        )
        # when we get the new schema of table version 1
        tv = metastore.get_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            catalog=self.catalog,
        )
        # expect it to be equivalent to the old schema (including metadata)
        assert tv.schema.equivalent_to(old_schema, True)
        # expect it to only have one schema in its evolution history
        assert len(tv.schemas) == 1
        assert tv.schemas[0].equivalent_to(old_schema, True)
        # expect the full table version to also be unchanged
        assert tv.equivalent_to(self.table_version)

    def test_update_table_version_schema_equivalent_schema_new_id(self):
        # given an update to only the schema ID of table version 1
        old_schema = self.table_version.schema
        new_schema = Schema.of(
            schema=old_schema.arrow,
            schema_id=old_schema.id + 1,
        )
        metastore.update_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            schema=new_schema,
            catalog=self.catalog,
        )
        # when we get the new schema of table version 1
        tv = metastore.get_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            catalog=self.catalog,
        )
        # expect it to be equivalent to the old schema (ignoring metadata)
        assert tv.schema.equivalent_to(old_schema)
        assert not tv.schema.equivalent_to(old_schema, True)
        assert tv.schema.id == new_schema.id != old_schema.id
        # expect it to have two schema in its evolution history
        assert len(tv.schemas) == 2
        assert tv.schemas[0].equivalent_to(old_schema, True)
        assert tv.schemas[0].id == old_schema.id
        assert tv.schemas[1].equivalent_to(old_schema)
        assert not tv.schemas[1].equivalent_to(old_schema, True)
        assert tv.schemas[1].id == new_schema.id != old_schema.id

    def test_update_table_version_partition_scheme(self):
        # given an update to the partition scheme of table version 1
        identity_transform = IdentityTransform.of()
        partition_keys = [
            PartitionKey.of(
                key=["some_string", "some_int32"],
                name="test_partition_key",
                field_id="test_field_id",
                transform=identity_transform,
            )
        ]
        new_scheme = PartitionScheme.of(
            keys=partition_keys,
            name="test_partition_scheme",
            scheme_id="test_partition_scheme_id_2",
        )
        metastore.update_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            partition_scheme=new_scheme,
            catalog=self.catalog,
        )
        # when we get the new partition scheme of table version 1
        tv = metastore.get_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            catalog=self.catalog,
        )
        # expect it to be equivalent to the expected scheme
        assert tv.partition_scheme.equivalent_to(new_scheme, True)
        assert tv.partition_scheme == new_scheme
        # expect the table version to have two schemes in its evolution history
        assert len(tv.partition_schemes) == 2
        old_scheme = self.table_version.partition_scheme
        assert tv.partition_schemes[0].equivalent_to(old_scheme, True)
        assert tv.partition_schemes[0] == old_scheme
        assert tv.partition_schemes[1].equivalent_to(new_scheme, True)
        assert tv.partition_schemes[1] == new_scheme
        # expect ONLY the partition scheme to be updated
        expected_tv = Metafile.update_for(self.table_version)
        expected_tv.partition_scheme = new_scheme
        expected_tv.partition_schemes = [old_scheme, new_scheme]
        assert tv.equivalent_to(expected_tv)

    def test_update_table_version_partition_scheme_same_id_fails(self):
        # given an update to table version 1 partition scheme using the same ID
        identity_transform = IdentityTransform.of()
        partition_keys = [
            PartitionKey.of(
                key=["some_string", "some_int32"],
                name="test_partition_key",
                field_id="test_field_id",
                transform=identity_transform,
            )
        ]
        new_scheme = PartitionScheme.of(
            keys=partition_keys,
            name="new_partition_scheme_name",
            scheme_id="test_partition_scheme_id",
        )
        # when we try to update the partition scheme
        # expect an error to be raised
        with pytest.raises(ValueError):
            metastore.update_table_version(
                namespace=self.table.namespace,
                table_name=self.table.table_name,
                table_version=self.table_version.table_version,
                partition_scheme=new_scheme,
                catalog=self.catalog,
            )

    def test_update_table_version_partition_scheme_equivalent_scheme_noop(self):
        # given a noop update to the partition scheme of table version 1
        old_scheme = self.table_version.partition_scheme
        new_scheme = copy.deepcopy(old_scheme)
        metastore.update_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            partition_scheme=new_scheme,
            catalog=self.catalog,
        )
        # when we get the new partition scheme of table version 1
        tv = metastore.get_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            catalog=self.catalog,
        )
        # expect it to be equal the old scheme (including identifiers)
        assert tv.partition_scheme.equivalent_to(old_scheme, True)
        assert tv.partition_scheme == old_scheme
        # expect it to only have one scheme in its evolution history
        assert len(tv.partition_schemes) == 1
        assert tv.partition_schemes[0].equivalent_to(old_scheme, True)
        assert tv.partition_schemes[0] == old_scheme
        # expect the full table version to also be unchanged
        assert tv.equivalent_to(self.table_version)

    def test_update_table_version_partition_scheme_equivalent_scheme_new_id(self):
        # given an update to only the partition scheme ID of table version 1
        old_scheme = self.table_version.partition_scheme
        new_scheme = PartitionScheme.of(
            keys=copy.deepcopy(old_scheme.keys),
            name=old_scheme.name,
            scheme_id=old_scheme.id + "_2",
        )
        metastore.update_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            partition_scheme=new_scheme,
            catalog=self.catalog,
        )
        # when we get table version 1
        tv = metastore.get_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            catalog=self.catalog,
        )
        # expect it to be equivalent to the old scheme (ignoring identifiers)
        assert tv.partition_scheme.equivalent_to(old_scheme, False)
        assert not tv.partition_scheme.equivalent_to(old_scheme, True)
        # expect it to have two schemes in its evolution history
        assert len(tv.partition_schemes) == 2
        assert tv.partition_schemes[0].equivalent_to(old_scheme, True)
        assert tv.partition_schemes[0].id == old_scheme.id != new_scheme.id
        assert tv.partition_schemes[1].equivalent_to(old_scheme)
        assert not tv.partition_schemes[1].equivalent_to(old_scheme, True)
        assert tv.partition_schemes[1].id == new_scheme.id != old_scheme.id

    def test_update_table_version_partition_scheme_equivalent_scheme_new_name(self):
        # given an update to the partition scheme name & ID of table version 1
        old_scheme = self.table_version.partition_scheme
        new_scheme = PartitionScheme.of(
            keys=copy.deepcopy(old_scheme.keys),
            name=old_scheme.name + "_2",
            scheme_id=old_scheme.id + "_2",
        )
        metastore.update_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            partition_scheme=new_scheme,
            catalog=self.catalog,
        )
        # when we get table version 1
        tv = metastore.get_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            catalog=self.catalog,
        )
        # expect it to be equivalent to the old scheme (ignoring identifiers)
        assert tv.partition_scheme.equivalent_to(old_scheme, False)
        assert tv.partition_scheme.id == new_scheme.id != old_scheme.id
        assert tv.partition_scheme.name == new_scheme.name != old_scheme.name
        assert not tv.partition_scheme.equivalent_to(old_scheme, True)
        # expect it to have two schemes in its evolution history
        assert len(tv.partition_schemes) == 2
        assert tv.partition_schemes[0].equivalent_to(old_scheme, True)
        assert tv.partition_schemes[0].id == old_scheme.id
        assert tv.partition_schemes[1].equivalent_to(old_scheme)
        assert not tv.partition_schemes[1].equivalent_to(old_scheme, True)
        assert tv.partition_schemes[1].id == new_scheme.id != old_scheme.id
        assert tv.partition_schemes[1].name == new_scheme.name != old_scheme.name

    def test_update_table_version_sort_scheme(self):
        # given an update to the sort scheme of table version 1
        sort_keys = [
            SortKey.of(
                key=["some_int32"],
                transform=IdentityTransform.of(),
            )
        ]
        new_scheme = SortScheme.of(
            keys=sort_keys,
            name="test_sort_scheme",
            scheme_id="test_sort_scheme_id_2",
        )
        metastore.update_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            sort_keys=new_scheme,
            catalog=self.catalog,
        )
        # when we get the new sort scheme of table version 1
        tv = metastore.get_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            catalog=self.catalog,
        )
        # expect it to be equivalent to the expected scheme
        assert tv.sort_scheme.equivalent_to(new_scheme, True)
        assert tv.sort_scheme == new_scheme
        # expect the table version to have two schemes in its evolution history
        assert len(tv.sort_schemes) == 2
        old_scheme = self.table_version.sort_scheme
        assert tv.sort_schemes[0].equivalent_to(old_scheme, True)
        assert tv.sort_schemes[0] == old_scheme
        assert tv.sort_schemes[1].equivalent_to(new_scheme, True)
        assert tv.sort_schemes[1] == new_scheme
        # expect ONLY the sort scheme to be updated
        expected_tv = Metafile.update_for(self.table_version)
        expected_tv.sort_scheme = new_scheme
        expected_tv.sort_schemes = [old_scheme, new_scheme]
        assert tv.equivalent_to(expected_tv)

    def test_update_table_version_sort_scheme_same_id_fails(self):
        # given an update to table version 1 sort scheme using the same ID
        sort_keys = [
            SortKey.of(
                key=["some_int32"],
                transform=IdentityTransform.of(),
            )
        ]
        new_scheme = SortScheme.of(
            keys=sort_keys,
            name="new_sort_scheme_name",
            scheme_id="test_sort_scheme_id",
        )
        # when we try to update the sort scheme
        # expect an error to be raised
        with pytest.raises(ValueError):
            metastore.update_table_version(
                namespace=self.table.namespace,
                table_name=self.table.table_name,
                table_version=self.table_version.table_version,
                sort_keys=new_scheme,
                catalog=self.catalog,
            )

    def test_update_table_version_sort_scheme_equivalent_scheme_noop(self):
        # given a noop update to the sort scheme of table version 1
        old_scheme = self.table_version.sort_scheme
        new_scheme = copy.deepcopy(old_scheme)
        metastore.update_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            sort_keys=new_scheme,
            catalog=self.catalog,
        )
        # when we get the new sort scheme of table version 1
        tv = metastore.get_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            catalog=self.catalog,
        )
        # expect it to be equal the old scheme (including identifiers)
        assert tv.sort_scheme.equivalent_to(old_scheme, True)
        assert tv.sort_scheme == old_scheme
        # expect it to only have one scheme in its evolution history
        assert len(tv.sort_schemes) == 1
        assert tv.sort_schemes[0].equivalent_to(old_scheme, True)
        assert tv.sort_schemes[0] == old_scheme
        # expect the full table version to also be unchanged
        assert tv.equivalent_to(self.table_version)

    def test_update_table_version_sort_scheme_equivalent_scheme_new_id(self):
        # given an update to only the sort scheme ID of table version 1
        old_scheme = self.table_version.sort_scheme
        new_scheme = SortScheme.of(
            keys=copy.deepcopy(old_scheme.keys),
            name=old_scheme.name,
            scheme_id=old_scheme.id + "_2",
        )
        metastore.update_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            sort_keys=new_scheme,
            catalog=self.catalog,
        )
        # when we get table version 1
        tv = metastore.get_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            catalog=self.catalog,
        )
        # expect it to be equivalent to the old scheme (ignoring identifiers)
        assert tv.sort_scheme.equivalent_to(old_scheme, False)
        assert not tv.sort_scheme.equivalent_to(old_scheme, True)
        # expect it to have two schemes in its evolution history
        assert len(tv.sort_schemes) == 2
        assert tv.sort_schemes[0].equivalent_to(old_scheme, True)
        assert tv.sort_schemes[0].id == old_scheme.id != new_scheme.id
        assert tv.sort_schemes[1].equivalent_to(old_scheme)
        assert not tv.sort_schemes[1].equivalent_to(old_scheme, True)
        assert tv.sort_schemes[1].id == new_scheme.id != old_scheme.id

    def test_update_table_version_sort_scheme_equivalent_scheme_new_name(self):
        # given an update to the sort scheme name & ID of table version 1
        old_scheme = self.table_version.sort_scheme
        new_scheme = SortScheme.of(
            keys=copy.deepcopy(old_scheme.keys),
            name=old_scheme.name + "_2",
            scheme_id=old_scheme.id + "_2",
        )
        metastore.update_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            sort_keys=new_scheme,
            catalog=self.catalog,
        )
        # when we get table version 1
        tv = metastore.get_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            catalog=self.catalog,
        )
        # expect it to be equivalent to the old scheme (ignoring identifiers)
        assert tv.sort_scheme.equivalent_to(old_scheme, False)
        assert tv.sort_scheme.id == new_scheme.id != old_scheme.id
        assert tv.sort_scheme.name == new_scheme.name != old_scheme.name
        assert not tv.sort_scheme.equivalent_to(old_scheme, True)
        # expect it to have two schemes in its evolution history
        assert len(tv.sort_schemes) == 2
        assert tv.sort_schemes[0].equivalent_to(old_scheme, True)
        assert tv.sort_schemes[0].id == old_scheme.id
        assert tv.sort_schemes[1].equivalent_to(old_scheme)
        assert not tv.sort_schemes[1].equivalent_to(old_scheme, True)
        assert tv.sort_schemes[1].id == new_scheme.id != old_scheme.id
        assert tv.sort_schemes[1].name == new_scheme.name != old_scheme.name

    def test_update_table_version_description(self):
        # given an update to the description of table version 1
        new_description = "new description"
        metastore.update_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            description=new_description,
            catalog=self.catalog,
        )
        # when we get table version 1
        tv = metastore.get_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            catalog=self.catalog,
        )
        # expect it to contain the new description
        assert tv.description == new_description != self.table_version.description
        # expect ONLY the description to be updated
        expected_tv = Metafile.update_for(self.table_version)
        expected_tv.description = new_description
        assert tv.equivalent_to(expected_tv)

    def test_update_table_version_description_empty(self):
        # given an update to create an empty description of table version 1
        new_description = ""
        metastore.update_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            description=new_description,
            catalog=self.catalog,
        )
        # when we get table version 1
        tv = metastore.get_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            catalog=self.catalog,
        )
        # expect it to contain the new description
        assert tv.description == new_description != self.table_version.description

    def test_update_table_version_description_noop(self):
        # given an attempt to set the description of table version 1
        metastore.update_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            description=None,
            catalog=self.catalog,
        )
        # when we get table version 1
        tv = metastore.get_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            catalog=self.catalog,
        )
        # expect it to contain the old description (None == noop)
        assert tv.description == self.table_version.description
        # expect the full table version to also be unchanged
        assert tv.equivalent_to(self.table_version)

    def test_update_table_version_properties(self):
        # given an update to the properties of table version 1
        new_properties = {"new_property_key": "new_property_value"}
        metastore.update_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            properties=new_properties,
            catalog=self.catalog,
        )
        # when we get table version 1
        tv = metastore.get_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            catalog=self.catalog,
        )
        # expect it to contain the new properties
        assert tv.properties == new_properties != self.table_version.properties
        # expect ONLY the properties to be updated
        expected_tv = Metafile.update_for(self.table_version)
        expected_tv.properties = new_properties
        assert tv.equivalent_to(expected_tv)

    def test_update_table_version_properties_empty(self):
        # given an update to leave table version 1 properties empty
        new_properties = {}
        metastore.update_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            properties=new_properties,
            catalog=self.catalog,
        )
        # when we get table version 1
        tv = metastore.get_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            catalog=self.catalog,
        )
        # expect it to contain the new properties
        assert tv.properties == new_properties != self.table_version.properties

    def test_update_table_version_properties_noop(self):
        # given an attempt to set the properties of table version 1
        metastore.update_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            properties=None,
            catalog=self.catalog,
        )
        # when we get table version 1
        tv = metastore.get_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            catalog=self.catalog,
        )
        # expect it to contain the old properties (None == noop)
        assert tv.properties == self.table_version.properties
        # expect the full table version to also be unchanged
        assert tv.equivalent_to(self.table_version)

    def test_get_latest_active_table_version(self):
        # given two table versions but no active table version
        # when we get the latest active table version
        tv = metastore.get_latest_active_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            catalog=self.catalog,
        )
        # expect it to be undefined
        assert tv is None
        # when we get the parent table
        table = metastore.get_table(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            catalog=self.catalog,
        )
        # expect its latest table version to be table version 2
        assert table.latest_table_version == self.table_version2.table_version
        # expect its latest active table version to be None
        assert table.latest_active_table_version is None
        # expect table attributes to be equal to the original parent table
        assert table.equivalent_to(self.table)

        # given an update to make table version 1 active
        metastore.update_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            lifecycle_state=LifecycleState.ACTIVE,
            catalog=self.catalog,
        )
        # when we get the latest active table version
        tv = metastore.get_latest_active_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            catalog=self.catalog,
        )
        # expect it to be table version 1
        active_table_version: TableVersion = Metafile.update_for(self.table_version)
        active_table_version.state = LifecycleState.ACTIVE
        assert tv.equivalent_to(active_table_version)
        # given an update to make table version 2 active
        metastore.update_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version2.table_version,
            lifecycle_state=LifecycleState.ACTIVE,
            catalog=self.catalog,
        )
        # when we get the latest active table version
        tv = metastore.get_latest_active_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            catalog=self.catalog,
        )
        # expect it to be table version 2
        active_table_version2: TableVersion = Metafile.update_for(self.table_version2)
        active_table_version2.state = LifecycleState.ACTIVE
        assert tv.equivalent_to(active_table_version2)

    def test_get_latest_active_table_version_bad_parent_locator(self):
        kwargs = {
            "namespace": self.table.namespace,
            "table_name": self.table.table_name,
        }
        for key in kwargs.keys():
            kwargs_copy = copy.copy(kwargs)
            # given a bad table version parent locator
            kwargs_copy[key] = "i_dont_exist"
            # when we get the latest active table version
            # expect an error to be raised
            with pytest.raises(ValueError):
                metastore.get_latest_active_table_version(
                    catalog=self.catalog,
                    **kwargs_copy,
                )

    def test_get_table_version(self):
        # given a previously created table version
        # when we explicitly get that table version by ID
        tv = metastore.get_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            catalog=self.catalog,
        )
        # expect the table version returned to be equivalent to the one created
        assert tv.equivalent_to(self.table_version)

    def test_get_table_version_not_exists(self):
        # given a previously created table
        # when we explicitly try to get table version whose ID doesn't exist
        tv = metastore.get_table_version(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version="v.999",
            catalog=self.catalog,
        )
        # expect nothing to be returned
        assert tv is None

    def test_get_table_version_bad_parent_locator(self):
        kwargs = {
            "namespace": self.table.namespace,
            "table_name": self.table.table_name,
        }
        for key in kwargs.keys():
            kwargs_copy = copy.copy(kwargs)
            # given a bad table version parent locator
            kwargs_copy[key] = "i_dont_exist"
            # when we try to explicitly get a table version by ID
            # expect result to be None
            assert (
                metastore.get_table_version(
                    table_version=self.table_version.table_version,
                    catalog=self.catalog,
                    **kwargs_copy,
                )
                is None
            )

    def test_table_version_exists(self):
        # given a previously created table version
        # when we check if that table version exists by ID
        # expect the check to pass
        assert metastore.table_version_exists(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version=self.table_version.table_version,
            catalog=self.catalog,
        )

    def test_table_version_not_exists(self):
        # given a previously created table
        # when we check if a non-existent table version ID exists
        # expect the check to fail
        assert not metastore.table_version_exists(
            namespace=self.table.namespace,
            table_name=self.table.table_name,
            table_version="v.999",
            catalog=self.catalog,
        )

    def test_table_version_exists_bad_parent_locator(self):
        kwargs = {
            "namespace": self.table.namespace,
            "table_name": self.table.table_name,
        }
        for key in kwargs.keys():
            kwargs_copy = copy.copy(kwargs)
            # given a bad table version parent locator
            kwargs_copy[key] = "i_dont_exist"
            # when we try to explicitly check if a table version exists by ID
            # expect empty results
            assert not metastore.table_version_exists(
                table_version=self.table_version.table_version,
                catalog=self.catalog,
                **kwargs_copy,
            )

    def test_creation_fails_if_already_exists(self):
        # given an existing table version
        # when we try to create a table version with the same ID
        # expect an error to be raised
        with pytest.raises(ValueError):
            metastore.create_table_version(
                namespace=self.table.namespace,
                table_name=self.table.table_name,
                table_version=self.table_version.table_version,
                catalog=self.catalog,
            )


class TestStream:
    @classmethod
    def setup_method(cls):
        cls.tmpdir = tempfile.mkdtemp()
        cls.catalog = CatalogProperties(root=cls.tmpdir)
        metastore.create_namespace(
            "test_stream_ns",
            catalog=cls.catalog,
        )
        # Create a table version.
        # This call should automatically create a default DeltaCAT stream.
        cls.table, cls.tv, cls.stream = metastore.create_table_version(
            namespace="test_stream_ns",
            table_name="mystreamtable",
            table_version="v.1",
            catalog=cls.catalog,
        )
        # Retrieve the auto-created default stream.
        cls.default_stream = metastore.get_stream(
            namespace="test_stream_ns",
            table_name="mystreamtable",
            table_version="v.1",
            catalog=cls.catalog,
        )
        # Ensure that the default stream was auto-created.
        assert cls.default_stream is not None, "Default stream not found."
        assert cls.default_stream.equivalent_to(cls.stream)

    @classmethod
    def teardown_method(cls):
        shutil.rmtree(cls.tmpdir)

    def test_list_streams(self):
        list_result = metastore.list_streams(
            "test_stream_ns",
            "mystreamtable",
            "v.1",
            catalog=self.catalog,
        )
        streams = list_result.all_items()
        # We expect exactly one stream (the default "deltacat" stream).
        assert len(streams) == 1
        assert streams[0].equivalent_to(self.default_stream)

    def test_stream_exists(self):
        exists = metastore.stream_exists(
            namespace="test_stream_ns",
            table_name="mystreamtable",
            table_version="v.1",
            catalog=self.catalog,
        )
        assert exists

    def test_stream_not_exists(self):
        exists = metastore.stream_exists(
            namespace="test_stream_ns",
            table_name="mystreamtable",
            table_version="v.1",
            stream_format=StreamFormat.ICEBERG,
            catalog=self.catalog,
        )
        assert not exists

    def test_stream_exists_bad_parent_locator(self):
        kwargs = {
            "namespace": "test_stream_ns",
            "table_name": "mystreamtable",
            "table_version": "v.1",
        }
        for key in kwargs.keys():
            kwargs_copy = copy.copy(kwargs)
            # table version format must be v.N to not raise a ValueError
            kwargs_copy[key] = "i_dont_exist" if key != "table_version" else "v.1000"
            assert not metastore.stream_exists(
                catalog=self.catalog,
                **kwargs_copy,
            )

    def test_list_streams_bad_parent_locator(self):
        kwargs = {
            "namespace": "test_stream_ns",
            "table_name": "mystreamtable",
            "table_version": "v.1",
        }
        for key in kwargs.keys():
            kwargs_copy = copy.copy(kwargs)
            kwargs_copy[key] = "i_dont_exist"
            with pytest.raises(ValueError):
                metastore.list_streams(
                    catalog=self.catalog,
                    **kwargs_copy,
                )

    def test_get_stream(self):
        stream = metastore.get_stream(
            namespace="test_stream_ns",
            table_name="mystreamtable",
            table_version="v.1",
            catalog=self.catalog,
        )
        assert stream.equivalent_to(self.default_stream)

    def test_get_stream_bad_parent_locator(self):
        kwargs = {
            "namespace": "test_stream_ns",
            "table_name": "mystreamtable",
            "table_version": "v.1",
        }
        for key in kwargs.keys():
            kwargs_copy = copy.copy(kwargs)
            # table version format must be v.N to not raise a ValueError
            kwargs_copy[key] = "i_dont_exist" if key != "table_version" else "v.1000"
            assert (
                metastore.get_stream(
                    catalog=self.catalog,
                    **kwargs_copy,
                )
                is None
            )

    def test_get_missing_stream(self):
        stream = metastore.get_stream(
            namespace="test_stream_ns",
            table_name="mystreamtable",
            table_version="v.1",
            stream_format=StreamFormat.ICEBERG,
            catalog=self.catalog,
        )
        assert stream is None

    def test_list_stream_partitions_empty(self):
        # no partitions yet
        list_result = metastore.list_stream_partitions(
            self.default_stream,
            catalog=self.catalog,
        )
        partitions = list_result.all_items()
        assert len(partitions) == 0

    def test_delete_stream(self):
        # Given a directive to delete the default stream
        metastore.delete_stream(
            namespace="test_stream_ns",
            table_name="mystreamtable",
            table_version="v.1",
            catalog=self.catalog,
        )
        # When we try to get the last committed stream
        stream = metastore.get_stream(
            namespace="test_stream_ns",
            table_name="mystreamtable",
            table_version="v.1",
            catalog=self.catalog,
        )
        # Expect nothing to be returned
        assert stream is None

        # Even when we try to get the last committed stream by ID
        stream = metastore.get_stream_by_id(
            table_version_locator=TableVersionLocator.at(
                namespace="test_stream_ns",
                table_name="mystreamtable",
                table_version="v.1",
            ),
            stream_id=self.default_stream.id,
            catalog=self.catalog,
        )
        # Expect nothing to be returned
        assert stream is None
        # TODO(pdames): Add new getter method for deleted but not GC'd streams?

    def test_delete_missing_stream(self):
        with pytest.raises(ValueError):
            metastore.delete_stream(
                namespace="test_stream_ns",
                table_name="mystreamtable",
                table_version="v.1",
                stream_format=StreamFormat.ICEBERG,
                catalog=self.catalog,
            )

    def test_delete_stream_bad_parent_locator(self):
        kwargs = {
            "namespace": "test_stream_ns",
            "table_name": "mystreamtable",
            "table_version": "v.1",
        }
        for key in kwargs.keys():
            kwargs_copy = copy.copy(kwargs)
            kwargs_copy[key] = "i_dont_exist"
            with pytest.raises(ValueError):
                metastore.delete_stream(
                    catalog=self.catalog,
                    **kwargs_copy,
                )

    def test_stage_and_commit_stream_replacement(self):
        # Given a staged stream that overwrites the default stream
        staged_stream = metastore.stage_stream(
            namespace="test_stream_ns",
            table_name="mystreamtable",
            table_version="v.1",
            catalog=self.catalog,
        )
        # When that staged stream is retrieved by ID
        fetched_stream = metastore.get_stream_by_id(
            table_version_locator=TableVersionLocator.at(
                namespace="test_stream_ns",
                table_name="mystreamtable",
                table_version="v.1",
            ),
            stream_id=staged_stream.stream_id,
            catalog=self.catalog,
        )
        # Ensure that it is equivalent to the stream we staged
        assert fetched_stream.id == staged_stream.id == fetched_stream.stream_id
        assert fetched_stream.equivalent_to(staged_stream)
        # Also ensure that the last committed deltacat stream returned is
        # NOT the staged stream, but the committed default stream.
        fetched_stream = metastore.get_stream(
            namespace="test_stream_ns",
            table_name="mystreamtable",
            table_version="v.1",
            catalog=self.catalog,
        )
        assert fetched_stream.id == self.default_stream.id == fetched_stream.stream_id
        assert fetched_stream.equivalent_to(self.default_stream)
        # Given a committed stream that replaces the default stream
        committed_stream = metastore.commit_stream(
            stream=staged_stream,
            catalog=self.catalog,
        )
        # When the last committed stream is retrieved
        fetched_stream = metastore.get_stream(
            namespace="test_stream_ns",
            table_name="mystreamtable",
            table_version="v.1",
            catalog=self.catalog,
        )
        # Ensure that it is equivalent to the stream we committed
        assert fetched_stream.id == committed_stream.id == fetched_stream.stream_id
        assert fetched_stream.equivalent_to(committed_stream)
        list_result = metastore.list_streams(
            "test_stream_ns",
            "mystreamtable",
            "v.1",
            catalog=self.catalog,
        )
        streams = list_result.all_items()
        # This will list the default stream and the newly committed stream
        for stream in streams:
            if stream.id == committed_stream.id:
                assert stream.equivalent_to(committed_stream)
            else:
                deprecated_default_stream: Stream = Metafile.update_for(
                    self.default_stream
                )
                deprecated_default_stream.state = CommitState.DEPRECATED
                assert stream.equivalent_to(deprecated_default_stream)
        assert len(streams) == 2



import deltacat.compute.converter.utils.iceberg_columns as sc
import daft


def download_data_table_and_append_iceberg_columns(
    file, columns_to_download, additional_columns_to_append, sequence_number
):
    # TODO; add S3 client kwargs
    table = download_parquet_with_daft_hash_applied(
        identify_columns=columns_to_download, file=file, s3_client_kwargs={}
    )
    if sc._FILE_PATH_COLUMN_NAME in additional_columns_to_append:
        table = sc.append_file_path_column(table, file.file_path)
    if sc._ORDERED_RECORD_IDX_COLUMN_NAME in additional_columns_to_append:
        record_idx_iterator = iter(range(len(table)))
        table = sc.append_record_idx_col(table, record_idx_iterator)
    return table


def download_parquet_with_daft_hash_applied(
    identify_columns, file, s3_client_kwargs, **kwargs
):
    from daft import TimeUnit

    # TODO: Add correct read kwargs as in:
    #  https://github.com/ray-project/deltacat/blob/383855a4044e4dfe03cf36d7738359d512a517b4/deltacat/utils/daft.py#L97

    coerce_int96_timestamp_unit = TimeUnit.from_str(
        kwargs.get("coerce_int96_timestamp_unit", "ms")
    )

    from deltacat.utils.daft import _get_s3_io_config

    # TODO: Use Daft SHA1 hash instead to minimize probably of data corruption
    io_config = _get_s3_io_config(s3_client_kwargs=s3_client_kwargs)
    df = daft.read_parquet(
        path=file.file_path,
        io_config=io_config,
        coerce_int96_timestamp_unit=coerce_int96_timestamp_unit,
    )
    df = df.select(daft.col(identify_columns[0]).hash())
    arrow_table = df.to_arrow()
    return arrow_table
